import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nptyping import NDArray
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MEnum, Quantity, Section, SubSection, SectionProxy,
    Reference, derived)
from nomad.datamodel.metainfo.simulation.calculation import Calculation
from nomad.datamodel.metainfo.simulation.run import Run
from nomad.datamodel.metainfo.simulation.system import System, Atoms
from .common import FastAccess


class Interface(MSection):
    '''
    Section containing results of an interface (stacking fault, gamma surface, etc.)
    workflow.
    '''

    m_def = Section(validate=False)

    energy_extrinsic_stacking_fault = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule / m **2',
        description='''
        Value of the relaxed extrinsic stacking fault energy per unit area.
        ''')

    energy_intrinsic_stacking_fault = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule / m **2',
        description='''
        Value of the relaxed intrinsic stacking fault energy per unit area.
        ''')

    dimensionality = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Dimensionality of the property, i.e. 1 for stacking fault energy and 2 for gamma
        surface.
        ''')

    shift_direction = Quantity(
        type=str,
        shape=['dimensionality'],
        description='''
        shift direction of the two crystal parts to calculate the fault energy.
        ''')

    n_displacements = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of displacements in the shift to calculate the fault energy.
        ''')

    displacement_fraction = Quantity(
        type=np.dtype(np.float64),
        shape=['dimensionality', 'n_displacements'],
        description='''
        Relative displacements of the two crystal parts along the direction indicated by
        shift_direction.
        ''')

    energy_fault_plane = Quantity(
        type=np.dtype(np.float64),
        shape=['n_displacements'],
        unit='joule / m ** 2',
        description='''
        Value of the relaxed excess energy per unit area for each displacement.
        ''')

    gamma_surface = Quantity(
        type=np.dtype(np.float64),
        shape=['n_displacements', 'n_displacements'],
        unit='joule / m ** 2',
        description='''
        Value of the gamma surface, i.e. the excess energy per unit area calculated for
        each displacement.
        ''')

    slip_fraction = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Relative displacement between two crystal parts where the energy is maximum.
        ''')

    energy_unstable_stacking_fault = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule / m **2',
        description='''
        Value of the relaxed unstable stacking fault energy per unit area.
        ''')

    energy_unstable_twinning_fault = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule / m **2',
        description='''
        Value of the relaxed unstable twinning energy per unit area.
        ''')


class Raman(MSection):
    '''
    Section containing results of a Raman workflow.
    '''

    m_def = Section(validate=False)

    n_modes = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of evaluated vibrational modes.
        ''')

    n_atoms = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of atoms in the simulation cell.
        ''')

    frequencies = Quantity(
        type=np.dtype(np.float64),
        shape=['n_modes'],
        unit='1 / m',
        description='''
        Calculated value of the Raman frequencies.
        ''')


class MagneticOrdering(MSection):
    '''
    Section containing results of a magnetic ordering workflow.
    '''

    m_def = Section(validate=False)

    n_structures = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of magnetic structures evaluated.
        ''')

    n_atoms = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of atoms in the simulation cell.
        ''')

    labels = Quantity(
        type=str,
        shape=['n_structures'],
        description='''
        Labels corresponding to each magnetic structure.
        ''')

    energies = Quantity(
        type=np.dtype(np.float64),
        shape=['n_structures'],
        unit='joule',
        description='''
        Calculated value of the energies corresponding to each magnetic structure.
        ''')

    magnetic_moments = Quantity(
        type=np.dtype(np.float64),
        shape=['n_structures', 'n_atoms'],
        unit='bohr_magneton',
        description='''
        Resulting atomic magnetic moments corresponding to each magnetic structure.
        ''')

    magnetic_deformations = Quantity(
        type=np.dtype(np.float64),
        shape=['n_structures'],
        unit='m',
        description='''
        Average atomic displacements after relaxation with respect to the non-magnetic
        case for each magnetic structure.
        ''')


class Adsorption(MSection):
    '''
    Section containing results of a surface adsorption workflow.
    '''

    m_def = Section(
        validate=False)

    n_sites = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of sites for which the adsorption energy is evaluated.
        ''')

    slab_miller_index = Quantity(
        type=np.dtype(np.int32),
        shape=[3],
        description='''
        Miller index of the slab.
        ''')

    slab = Quantity(
        type=str,
        shape=[],
        description='''
        Chemical formula of the slab.
        ''')

    adsorbate = Quantity(
        type=str,
        shape=[],
        description='''
        Chemical formula of the adsorbate molecule.
        ''')

    adsorption_sites = Quantity(
        type=np.dtype(np.float64),
        shape=['n_sites'],
        description='''
        Coordinates of the adsorption sites corresponding to a minimum energy.
        ''')

    adsorption_energies = Quantity(
        type=np.dtype(np.float64),
        shape=['n_sites'],
        unit='joule',
        description='''
        Calculated value of the adsorption energy corresponding to each site.
        ''')


class ConvexHull(MSection):
    '''
    Section containing results of a convex hull workflow.
    '''

    m_def = Section(
        validate=False)

    n_elements = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of elements for which the thermal stability is evaluated. This represents
        the dimensionality of the convex hull.
        ''')

    n_points = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of points for which the energies are evaluated.
        ''')

    compositions = Quantity(
        type=np.dtype(np.float64),
        shape=['n_points', 'n_elements'],
        description='''
        Normalized composition of the elements corresponding to each point for which the
        energies are evaluated.
        ''')

    references = Quantity(
        type=str,
        shape=['n_elements'],
        description='''
        Specifies the reference structure for each element.
        ''')

    energy_of_formation = Quantity(
        type=np.dtype(np.float64),
        shape=['n_points'],
        unit='joule',
        description='''
        Values of the heat of formation corresponding to each point.
        ''')

    energy_hulll = Quantity(
        type=np.dtype(np.float64),
        shape=['n_points'],
        unit='joule',
        description='''
        Values of the energy above the convex hull corresponding to each point.
        ''')


class NudgedElasticBand(MSection):
    '''
    Section containing results of a nudged-elastic band workflow.
    '''

    m_def = Section(
        validate=False)

    method = Quantity(
        type=str,
        shape=[],
        description='''
        Specifies the method used in calculating the minumum energy path. Can be one of
        standard, improved_tangeant, full_spring_force, spline_interpolation, string.
        ''')

    climbing_image = Quantity(
        type=bool,
        shape=[],
        description='''
        Indicates if climbing image is used.
        ''')

    solid_state = Quantity(
        type=bool,
        shape=[],
        description='''
        Indicates if solid state nudged-elastic band calculation is performed.
        ''')

    optimizer = Quantity(
        type=str,
        shape=[],
        description='''
        Specifies the method used in energy minimization.
        ''')

    n_images = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of images used in the calculation.
        ''')

    spring_constants = Quantity(
        type=np.dtype(np.float64),
        shape=['n_images'],
        unit='newton',
        description='''
        Spring constants for each spring.
        ''')

    energy_barrier = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Maximum value of the calculated energy barrier.
        ''')

    force_maximum = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='newton',
        description='''
        Maximum force along the minimum energy path.
        ''')


class EOSFit(MSection):
    '''
    Section containing results of an equation of state fit.
    '''

    m_def = Section(validate=False)

    function_name = Quantity(
        type=str,
        shape=[],
        description='''
        Specifies the function used to perform the fitting of the volume-energy data. Value
        can be one of birch_euler, birch_lagrange, birch_murnaghan, mie_gruneisen,
        murnaghan, pack_evans_james, poirier_tarantola, tait, vinet.
        ''')

    fitted_energies = Quantity(
        type=np.dtype(np.float64),
        shape=['n_points'],
        unit='joule',
        description='''
        Array of the fitted energies corresponding to each volume.
        ''')

    bulk_modulus = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='pascal',
        description='''
        Calculated value of the bulk modulus by fitting the volume-energy data.
        ''')

    bulk_modulus_derivative = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Calculated value of the pressure derivative of the bulk modulus.
        ''')

    equilibrium_volume = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='m ** 3',
        description='''
        Calculated value of the equilibrium volume by fitting the volume-energy data.
        ''')

    equilibrium_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Calculated value of the equilibrium energy by fitting the volume-energy data.
        ''')

    rms_error = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Root-mean squared value of the error in the fitting.
        ''')


class EquationOfState(MSection):
    '''
    Section containing results of an equation of state workflow.
    '''

    m_def = Section(validate=False)

    n_points = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of volume-energy pairs in data.
        ''')

    volumes = Quantity(
        type=np.dtype(np.float64),
        shape=['n_points'],
        unit='m ** 3',
        description='''
        Array of volumes per atom for which the energies are evaluated.
        ''')

    energies = Quantity(
        type=np.dtype(np.float64),
        shape=['n_points'],
        unit='joule',
        description='''
        Array of energies corresponding to each volume.
        ''')

    eos_fit = SubSection(sub_section=EOSFit.m_def, repeats=True)


class DebyeModel(MSection):
    '''
    Section containing results of an debye-model workflow.
    '''

    m_def = Section(
        validate=False)

    n_temperatures = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of temperature evaluations.
        ''')

    thermal_conductivity = Quantity(
        type=np.dtype(np.float64),
        shape=['n_temperatures'],
        unit='watt / meter * kelvin',
        description='''
        Calculated value of the thermal conductity.
        ''')

    debye_temperature = Quantity(
        type=np.dtype(np.float64),
        shape=['n_temperatures'],
        unit='kelvin',
        description='''
        Calculated value of the Debye temperature.
        ''')

    gruneisen_parameter = Quantity(
        type=np.dtype(np.float64),
        shape=['n_temperatures'],
        description='''
        Calculated value of the Gruneisen parameter.
        ''')

    thermal_expansion = Quantity(
        type=np.dtype(np.float64),
        shape=['n_temperatures'],
        unit='1 / kelvin',
        description='''
        Calculated value of the thermal expansion coefficient.
        ''')

    bulk_modulus_static = Quantity(
        type=np.dtype(np.float64),
        shape=['n_temperatures'],
        unit='pascal',
        description='''
        Calculated value of the static bulk modulus.
        ''')

    bulk_modulus_isothermal = Quantity(
        type=np.dtype(np.float64),
        shape=['n_temperatures'],
        unit='pascal',
        description='''
        Calculated value of the static bulk modulus.
        ''')


class GeometryOptimization(MSection):
    '''
    Section containing the results of a geometry_optimization workflow.
    '''

    m_def = Section(
        validate=False)

    type = Quantity(
        type=str,
        shape=[],
        description='''
        The type of geometry optimization can either be ionic, cell_shape, cell_volume.
        ''')

    method = Quantity(
        type=str,
        shape=[],
        description='''
        The method used for geometry optimization.
        ''')

    convergence_tolerance_energy_difference = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        The input energy difference tolerance criterion.
        ''')

    convergence_tolerance_force_maximum = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='newton',
        description='''
        The input maximum net force tolerance criterion.
        ''')

    convergence_tolerance_displacement_maximum = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        The input maximum displacement tolerance criterion.
        ''')

    final_energy_difference = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        The difference in the energy_total between the last two steps during
        optimization.
        ''')

    final_force_maximum = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='newton',
        description='''
        The maximum net force in the last optimization step.
        ''')

    final_displacement_maximum = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        The maximum displacement in the last optimization step with respect to previous.
        ''')

    optimization_steps = Quantity(
        type=int,
        shape=[],
        description='''
        Number of optimization steps.
        ''')

    energies = Quantity(
        type=np.dtype(np.float64),
        unit='joule',
        shape=["optimization_steps"],
        description='''
        List of energy_total values gathered from the single configuration
        calculations that are a part of the optimization trajectory.
        ''')

    is_converged_geometry = Quantity(
        type=bool,
        shape=[],
        description='''
        Indicates if the geometry convergence criteria were fulfilled.
        ''')


class Phonon(MSection):
    '''
    Section containing the results of a phonon workflow.
    '''

    m_def = Section(
        validate=False)

    force_calculator = Quantity(
        type=str,
        shape=[],
        description='''
        Name of the program used to calculate the forces.
        ''')

    mesh_density = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='1 / meter ** 3',
        description='''
        Density of the k-mesh for sampling.
        ''')

    n_imaginary_frequencies = Quantity(
        type=int,
        shape=[],
        description='''
        Number of modes with imaginary frequencies.
        ''')

    random_displacements = Quantity(
        type=bool,
        shape=[],
        description='''
        Identifies if displacements are made randomly.
        ''')

    with_non_analytic_correction = Quantity(
        type=bool,
        shape=[],
        description='''
        Identifies if non-analytical term corrections are applied to dynamical matrix.
        ''')

    with_grueneisen_parameters = Quantity(
        type=bool,
        shape=[],
        description='''
        Identifies if Grueneisen parameters are calculated.
        ''')

    n_bands = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of phonon bands.
        ''')

    n_qpoints = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of q points for which phonon properties are evaluated.
        ''')

    qpoints = Quantity(
        type=np.dtype(np.float64),
        shape=['n_qpoints', 3],
        description='''
        Value of the qpoints.
        ''')

    group_velocity = Quantity(
        type=np.dtype(np.float64),
        shape=['n_qpoints', 'n_bands', 3],
        unit='meter / second',
        description='''
        Calculated value of the group velocity at each qpoint.
        ''')

    n_displacements = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of independent displacements.
        ''')

    n_atoms = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of atoms in the simulation cell.
        ''')

    displacements = Quantity(
        type=np.dtype(np.float64),
        shape=['n_displacements', 'n_atoms', 3],
        unit='meter',
        description='''
        Value of the displacements applied to each atom in the simulation cell.
        ''')


class StrainDiagrams(MSection):
    '''
    Section containing the information regarding the elastic strains.
    '''

    m_def = Section(
        validate=False)

    type = Quantity(
        type=str,
        shape=[],
        description='''
        Kind of strain diagram. Possible values are: energy; cross-validation (cross-
        validation error); d2E (second derivative of the energy wrt the strain)
        ''')

    n_eta = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of strain values used in the strain diagram
        ''')

    n_deformations = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of deformations.
        ''')

    value = Quantity(
        type=np.dtype(np.float64),
        shape=['n_deformations', 'n_eta'],
        description='''
        Values of the energy(units:J)/d2E(units:Pa)/cross-validation (depending on the
        value of strain_diagram_type)
        ''')

    eta = Quantity(
        type=np.dtype(np.float64),
        shape=['n_deformations', 'n_eta'],
        description='''
        eta values used the strain diagrams
        ''')

    stress_voigt_component = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Voigt component corresponding to the strain diagram
        ''')

    polynomial_fit_order = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Order of the polynomial fit
        ''')


class Elastic(MSection):
    '''
    Section containing the results of an elastic workflow.
    '''

    m_def = Section(
        validate=False)

    energy_stress_calculator = Quantity(
        type=str,
        shape=[],
        description='''
        Name of program used to calculate energy or stress.
        ''')

    calculation_method = Quantity(
        type=str,
        shape=[],
        description='''
        Method used to calculate elastic constants, can either be energy or stress.
        ''')

    elastic_constants_order = Quantity(
        type=int,
        shape=[],
        description='''
        Order of the calculated elastic constants.
        ''')

    n_deformations = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of deformed structures used to calculate the elastic constants. This is
        determined by the symmetry of the crystal.
        ''')

    deformation_types = Quantity(
        type=np.dtype('U'),
        shape=['n_deformations', 6],
        description='''
        deformation types
        ''')

    n_strains = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        number of equally spaced strains applied to each deformed structure, which are
        generated between the maximum negative strain and the maximum positive one.
        ''')

    is_mechanically_stable = Quantity(
        type=bool,
        shape=[],
        description='''
        Indicates if structure is mechanically stable from the calculated values of the
        elastic constants.
        ''')

    fitting_error_maximum = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Maximum error in polynomial fit.
        ''')

    strain_maximum = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Maximum strain applied to crystal.
        ''')

    elastic_constants_notation_matrix_second_order = Quantity(
        type=np.dtype('U'),
        shape=[6, 6],
        description='''
        Symmetry of the second-order elastic constant matrix in Voigt notation
        ''')

    elastic_constants_matrix_second_order = Quantity(
        type=np.dtype(np.float64),
        shape=[6, 6],
        unit='pascal',
        description='''
        2nd order elastic constant (stiffness) matrix in pascals
        ''')

    elastic_constants_matrix_third_order = Quantity(
        type=np.dtype(np.float64),
        shape=[6, 6, 6],
        unit='pascal',
        description='''
        3rd order elastic constant (stiffness) matrix in pascals
        ''')

    compliance_matrix_second_order = Quantity(
        type=np.dtype(np.float64),
        shape=[6, 6],
        unit='1 / pascal',
        description='''
        Elastic compliance matrix in 1/GPa
        ''')

    elastic_constants_gradient_matrix_second_order = Quantity(
        type=np.dtype(np.float64),
        shape=[18, 18],
        unit='newton',
        description='''
        gradient of the 2nd order elastic constant (stiffness) matrix in newton
        ''')

    bulk_modulus_voigt = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='pascal',
        description='''
        Voigt bulk modulus
        ''')

    shear_modulus_voigt = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='pascal',
        description='''
        Voigt shear modulus
        ''')

    bulk_modulus_reuss = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='pascal',
        description='''
        Reuss bulk modulus
        ''')

    shear_modulus_reuss = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='pascal',
        description='''
        Reuss shear modulus
        ''')

    bulk_modulus_hill = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='pascal',
        description='''
        Hill bulk modulus
        ''')

    shear_modulus_hill = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='pascal',
        description='''
        Hill shear modulus
        ''')

    young_modulus_voigt = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='pascal',
        description='''
        Voigt Young modulus
        ''')

    poisson_ratio_voigt = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Voigt Poisson ratio
        ''')

    young_modulus_reuss = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='pascal',
        description='''
        Reuss Young modulus
        ''')

    poisson_ratio_reuss = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Reuss Poisson ratio
        ''')

    young_modulus_hill = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='pascal',
        description='''
        Hill Young modulus
        ''')

    poisson_ratio_hill = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Hill Poisson ratio
        ''')

    elastic_anisotropy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Elastic anisotropy
        ''')

    pugh_ratio_hill = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Pugh ratio defined as the ratio between the shear modulus and bulk modulus
        ''')

    debye_temperature = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='kelvin',
        description='''
        Debye temperature
        ''')

    speed_sound_transverse = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter / second',
        description='''
        Speed of sound along the transverse direction
        ''')

    speed_sound_longitudinal = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter / second',
        description='''
        Speed of sound along the longitudinal direction
        ''')

    speed_sound_average = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter / second',
        description='''
        Average speed of sound
        ''')

    eigenvalues_elastic = Quantity(
        type=np.dtype(np.float64),
        shape=[6],
        unit='pascal',
        description='''
        Eigenvalues of the stiffness matrix
        ''')

    strain_diagrams = SubSection(
        sub_section=StrainDiagrams.m_def,
        repeats=True)


class Decomposition(MSection):
    '''
    Section containing information about the system to which an unstable compound will
    decompose to.
    '''

    m_def = Section(validate=False)

    fraction = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Amount of the resulting system.
        ''')

    system_ref = Quantity(
        type=Reference(System.m_def),
        shape=[],
        description='''
        Reference to the resulting system.
        ''')

    formula = Quantity(
        type=str,
        shape=[],
        description='''
        Chemical formula of the resulting system.
        ''')


class Stability(MSection):
    '''
    Section containing information regarding the stability of the system.
    '''

    m_def = Section(validate=False)

    n_references = Quantity(
        type=int,
        shape=[],
        description='''
        Number of reference systems.
        ''')

    systems_ref = Quantity(
        type=Reference(System.m_def),
        shape=['n_references'],
        description='''
        References to the reference systems.
        ''')

    formation_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Calculated value of the formation energy of the compound.
        ''')

    delta_formation_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Energy with respect to the convex hull.
        ''')

    n_references = Quantity(
        type=int,
        shape=[],
        description='''
        Number of reference systems.
        ''')

    is_stable = Quantity(
        type=bool,
        shape=[],
        description='''
        Indicates if a compound is stable.
        ''')

    decomposition = SubSection(sub_section=Decomposition.m_def, repeats=True)


class Thermodynamics(MSection):
    '''
    Section containing the results of a thermodynamics workflow.
    '''

    m_def = Section(validate=False)

    n_values = Quantity(
        type=int,
        shape=[],
        description='''
        Number of thermodynamics property evaluations.
        ''')

    temperature = Quantity(
        type=np.dtype(np.float64),
        shape=['n_values'],
        unit='kelvin',
        description='''
        Specifies the temperatures at which properties such as the Helmholtz free energy
        are calculated.
        ''')

    pressure = Quantity(
        type=np.dtype(np.float64),
        shape=['n_values'],
        unit='pascal',
        description='''
        Array containing the values of the pressure (one third of the trace of the stress
        tensor) corresponding to each property evaluation.
        ''')

    helmholtz_free_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['n_values'],
        unit='joule',
        description='''
        Helmholtz free energy per unit cell at constant volume.
        ''')

    heat_capacity_c_p = Quantity(
        type=np.dtype(np.float64),
        shape=['n_values'],
        unit='joule / kelvin',
        description='''
        Heat capacity per cell unit at constant pressure.
        ''')

    heat_capacity_c_v = Quantity(
        type=np.dtype(np.float64),
        shape=['n_values'],
        unit='joule / kelvin',
        description='''
        Heat capacity per cell unit at constant volume.
        ''')

    @derived(
        type=np.dtype(np.float64),
        shape=['n_values'],
        unit='joule / (kelvin * kilogram)',
        description='''
        Specific heat capacity at constant volume.
        ''',
        cached=True
    )
    def heat_capacity_c_v_specific(self) -> NDArray:
        """Returns the specific heat capacity by dividing the heat capacity per
        cell with the mass of the atoms in the cell.
        """
        import nomad.atomutils
        workflow = self.m_parent
        system = workflow.calculations_ref[0].system_ref
        atomic_numbers = system.atoms.species
        mass_per_unit_cell = nomad.atomutils.get_summed_atomic_mass(atomic_numbers)
        heat_capacity = self.heat_capacity_c_v
        specific_heat_capacity = heat_capacity / mass_per_unit_cell

        return specific_heat_capacity.magnitude

    vibrational_free_energy_at_constant_volume = Quantity(
        type=np.dtype(np.float64),
        shape=['n_values'],
        unit='joule',
        description='''
        Holds the vibrational free energy per cell unit at constant volume.
        ''')

    @derived(
        type=np.dtype(np.float64),
        shape=['n_values'],
        unit='joule / kilogram',
        description='''
        Stores the specific vibrational free energy at constant volume.
        ''',
        cached=True
    )
    def vibrational_free_energy_at_constant_volume_specific(self) -> NDArray:
        import nomad.atomutils
        workflow = self.m_parent
        system = workflow.calculations_ref[0].system_ref
        atomic_numbers = system.atoms.species
        mass_per_unit_cell = nomad.atomutils.get_summed_atomic_mass(atomic_numbers)
        free_energy = self.vibrational_free_energy_at_constant_volume
        specific_free_energy = free_energy / mass_per_unit_cell

        return specific_free_energy.magnitude

    vibrational_free_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['n_values'],
        unit='joule',
        description='''
        Calculated value of the vibrational free energy, F_vib.
        ''')

    vibrational_internal_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['n_values'],
        unit='joule',
        description='''
        Calculated value of the vibrational internal energy, U_vib.
        ''')

    vibrational_entropy = Quantity(
        type=np.dtype(np.float64),
        shape=['n_values'],
        unit='joule / kelvin',
        description='''
        Calculated value of the vibrational entropy, S.
        ''')

    gibbs_free_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['n_values'],
        unit='joule',
        description='''
        Calculated value of the Gibbs free energy, G.
        ''')

    entropy = Quantity(
        type=np.dtype(np.float64),
        shape=['n_values'],
        unit='joule / kelvin',
        description='''
        Calculated value of the entropy.
        ''')

    internal_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['n_values'],
        unit='joule',
        description='''
        Calculated value of the internal energy, U.
        ''')

    stability = SubSection(sub_section=Stability.m_def, repeats=False)


class MolecularDynamics(MSection):
    '''
    Section containing results of molecular dynamics workflow.
    '''

    m_def = Section(
        validate=False)

    ensemble_type = Quantity(
        type=str,
        shape=[],
        description='''
        Ensemble type used.
        ''')

    timestep = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='s',
        description='''
        Time step in the simulation.
        ''')

    finished_normally = Quantity(
        type=bool,
        shape=[],
        description='''
        Indicates if calculation terminated normally.
        ''')

    with_trajectory = Quantity(
        type=bool,
        shape=[],
        description='''
        Indicates if calculation includes trajectory data.
        ''')

    with_thermodynamics = Quantity(
        type=bool,
        shape=[],
        description='''
        Indicates if calculation contains thermodynamic data.
        ''')


class SinglePoint(MSection):
    '''
    Section containing results of a single point workflow.
    '''

    m_def = Section(
        validate=False)

    method = Quantity(
        type=str,
        shape=[],
        description='''
        Calculation method used.
        ''')

    n_scf_steps = Quantity(
        type=int,
        shape=[],
        description='''
        Number of self-consistent steps in the calculation
        ''')

    final_scf_energy_difference = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        The difference in the energy between the last two scf steps.
        ''')

    is_converged = Quantity(
        type=bool,
        shape=[],
        description='''
        Indicates if the convergence criteria were fullfilled
        ''')

    with_density_of_states = Quantity(
        type=bool,
        shape=[],
        description='''
        Indicates if the calculation contains density of states data
        ''')

    with_bandstructure = Quantity(
        type=bool,
        shape=[],
        description='''
        Indicates if the calculation contains bandstructure data
        ''')

    with_eigenvalues = Quantity(
        type=bool,
        shape=[],
        description='''
        Indicates if the calculation contains eigenvalue data
        ''')

    with_volumetric_data = Quantity(
        type=bool,
        shape=[],
        description='''
        Indicates if the calculation contains volumetric data
        ''')

    with_excited_states = Quantity(
        type=bool,
        shape=[],
        description='''
        Indicates if the calculation contains excited states data
        ''')


class Task(MSection):
    '''
    Section defining a specific task in the workflow chain. It has an input and an output,
    both can either be a workflow or a calculation and their relation is noted in the
    description.
    '''

    m_def = Section(
        validate=False)

    input_workflow = Quantity(
        type=Reference(SectionProxy('Workflow')),
        shape=[],
        description='''
        Reference to the input workflow.
        ''')

    output_workflow = Quantity(
        type=Reference(SectionProxy('Workflow')),
        shape=[],
        description='''
        Reference to the output workflow.
        ''')

    input_calculation = Quantity(
        type=Reference(Calculation.m_def),
        shape=[],
        description='''
        Reference to the input calculation.
        ''')

    output_calculation = Quantity(
        type=Reference(Calculation.m_def),
        shape=[],
        description='''
        Reference to the output calculation.
        ''')

    description = Quantity(
        type=str,
        shape=[],
        description='''
        Descibes the relationship between the input and output workflows. For example, if
        a single_point workflow uses the relaxed structure from a geometry_optimization as
        an input, the description may be "relaxed structure from geometry_optimization
        used to calucalate single_point properties"
        ''')


class Workflow(MSection):
    '''
    Section containing the  results of a workflow.
    '''

    m_def = Section(
        validate=False)

    type = Quantity(
        type=MEnum([
            "single_point",
            "geometry_optimization",
            "phonon",
            "elastic",
            "molecular_dynamics",
            "debye_model",
            "equation_of_state",
            "nudged_elastic_band",
            "convex_hull",
            "adsorption",
            "magnetic_ordering",
            "raman",
            "interface",
            "thermodynamics"
        ]),
        shape=[],
        description='''
        The workflow type.
        ''')

    initial_structure = Quantity(
        type=Reference(Atoms.m_def),
        shape=[],
        description='''
        Starting structure for geometry optimization.
        ''')

    calculator = Quantity(
        type=str,
        shape=[],
        description='''
        Energy and force calculator.
        ''')

    calculation_result_ref = Quantity(
        type=Reference(Calculation.m_def),
        shape=[],
        description='''
        Reference to calculation result. In the case of geometry_optimization and
        molecular dynamics, this corresponds to the final step in the simulation. For the
        rest of the workflow types, it refers to the original system.
        ''',
        categories=[FastAccess])

    n_calculations = Quantity(
        type=int,
        shape=[],
        description='''
        Number of calculations in workflow
        ''')

    calculations_ref = Quantity(
        type=Reference(Calculation.m_def),
        shape=['n_calculations'],
        description='''
        List of references to each section single_configuration_calculation in the
        simulation.
        ''')

    run_ref = Quantity(
        type=Reference(Run.m_def),
        shape=[],
        description='''
        Links the section workflow to the section run that contains the calculations.
        ''',
        categories=[FastAccess])

    n_references = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
         Number of references to the current section workflow.
        ''')

    workflows_ref = Quantity(
        type=Reference(SectionProxy('Workflow')),
        shape=['n_references'],
        description='''
        Links the the current section to other workflow sections. Such a link is necessary
        for example between an Debye model that uses a the poisson ratio calculated
        from an elastic workflow.
        ''',
        categories=[FastAccess])

    task = SubSection(sub_section=Task.m_def, repeats=True)

    single_point = SubSection(
        sub_section=SinglePoint.m_def,
        # TODO determine if there is a need for this to be a repeating section
        # such as in the context of fhi-vibes single_point
        repeats=False,
        categories=[FastAccess])

    geometry_optimization = SubSection(
        sub_section=GeometryOptimization.m_def,
        repeats=False,
        categories=[FastAccess])

    phonon = SubSection(
        sub_section=Phonon.m_def,
        repeats=False,
        categories=[FastAccess])

    elastic = SubSection(
        sub_section=Elastic.m_def,
        repeats=False)

    molecular_dynamics = SubSection(
        sub_section=MolecularDynamics.m_def,
        repeats=False)

    debye_model = SubSection(
        sub_section=DebyeModel.m_def,
        repeats=False)

    equation_of_state = SubSection(
        sub_section=EquationOfState.m_def,
        repeats=False)

    nudged_elastic_band = SubSection(
        sub_section=NudgedElasticBand.m_def,
        repeats=False)

    convex_hull = SubSection(
        sub_section=ConvexHull.m_def,
        repeats=False)

    adsorption = SubSection(
        sub_section=Adsorption.m_def,
        repeats=False)

    magnetic_ordering = SubSection(
        sub_section=MagneticOrdering.m_def,
        repeats=False)

    raman = SubSection(
        sub_section=Raman.m_def,
        repeats=False)

    interface = SubSection(
        sub_section=Interface.m_def,
        repeats=False)

    thermodynamics = SubSection(
        sub_section=Thermodynamics.m_def,
        repeats=False)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import re
import logging
import json
from datetime import datetime
from ase.cell import Cell
from ase.atoms import Atoms as aseatoms
from ase.spacegroup import Spacegroup
import numpy as np

from nomad.units import ureg
from nomad.datamodel import EntryArchive
from nomad.client import api

from nomad.datamodel.metainfo.simulation.run import Run, Program
from nomad.datamodel.metainfo.simulation.system import System, Atoms
from nomad.datamodel.metainfo.simulation.method import Method, ForceField, Model
from nomad.datamodel.metainfo.simulation.calculation import (
    BandEnergies, BandStructure, Calculation, Energy, EnergyEntry, Thermodynamics, Stress, StressEntry)
from nomad.datamodel.metainfo.workflow import Phonon, Workflow, Elastic, Interface
from .metainfo import openkim  # pylint: disable=unused-import


class Converter:
    def __init__(self, entries, logger=None):
        self.entries = entries
        self.logger = logger if logger is not None else logging.getLogger('__name__')
        self.material = dict()

    @property
    def entries(self):
        return self._entries

    @entries.setter
    def entries(self, value):
        self._entries = value
        self.material = dict()
        self.archive = EntryArchive()

    def calculate_nomad_error(self, property_name, transform_function=None):
        property_map = {
            # TODO add other properties: phonon
            # challenge for phonon is alligning phonon frequency array or extracting
            # frequencies at high-symm k-points
            'workflow[-1].elastic.elastic_constants_matrix_second_order': 'mechanical',
            'run[-1].calculation[-1].band_structure_phonon[-1].segment[*].energies': 'vibrational',
        }
        if property_name not in property_map:
            return

        try:
            elements = list(set(self.archive.run[-1].system[-1].atoms.labels))
        except Exception:
            return

        if transform_function is None:
            transform_function = lambda archive, property_name: np.array(archive.m_xpath(property_name)).flatten()

        page_after_value = None
        nomad_data = []
        query = {
            'results.material.elements': elements,
            'results.material.n_elements': len(elements),
            'results.properties.available_properties': property_map.get(property_name),
        }
        if self.material.get('space_group_number') is not None:
            query['results.material.symmetry.space_group_number'] = self.material.get('space_group_number')
        elif self.material.get('structure_name') is not None:
            query['results.material.symmetry.structure_name'] = self.material.get('structure_name')
        while True:
            response = api.post('entries/archive/query', data=json.dumps({
                'owner': 'public',
                'query': query,
                'pagination': {
                    'page_size': 5,
                    'page_after_value': page_after_value
                },
                'required': {
                    'run': {
                        'calculation[-1]': '*',
                        'system[-1]': {
                            'atoms': '*'
                        }
                    },
                    'workflow': '*'
                }
            }))
            assert response.status_code == 200
            result = response.json()
            nomad_data.extend([data['archive'] for data in result['data']])
            page_after_value = result['pagination'].get('next_page_after_value')

            if len(result['data']) < 5 or page_after_value is None:
                break

        if len(nomad_data) == 0:
            return

        try:
            nomad_data = np.array([transform_function(EntryArchive().m_from_dict(data), property_name) for data in nomad_data])
            openkim_data = np.array(transform_function(self.archive, property_name))
        except Exception:
            self.logger.error('Error transforming data.')

        if np.shape(nomad_data[0]) != np.shape(openkim_data):
            self.logger.error('Incompatible shape of openkim data and nomad.')
            return

        # calculate rms error of openkim data wrt nomad
        sec_workflow = self.archive.workflow[-1]
        sec_workflow.x_openkim_property = property_name
        sec_workflow.x_openkim_n_nomad_data = len(nomad_data)
        sec_workflow.x_openkim_nomad_std = np.average(np.std(nomad_data, axis=0))
        sec_workflow.x_openkim_nomad_rms_error = np.sqrt(np.average((nomad_data - openkim_data)**2))
        # TODO add references to the nomad calculations / workflows

    def convert(self, filename='openkim_archive.json'):
        def get_value(entry, key, array=False, default=None):
            val = entry.get(key, [] if array else default)
            return [val] if array and not isinstance(val, list) else val

        def symmetrize_matrix(matrix):
            matrix = np.array(matrix)
            return matrix + matrix.T - np.diag(matrix.diagonal())

        def get_atoms(entry):
            symbols = entry.get('species.source-value', [])
            basis = entry.get('basis-atom-coordinates.source-value', [[0., 0., 0.]])
            cellpar = []
            for x in ['a', 'b', 'c']:
                value = entry.get(f'{x}.si-value', cellpar[0] if cellpar else 1)
                cellpar.append([value] if not isinstance(value, list) else value)
            cellpar = (cellpar * ureg.m).to('angstrom').magnitude

            # TODO are angles denoted by alpha, beta, gamma in openkim? can they be lists?
            alpha = entry.get('alpha.source-value', 90)
            beta = entry.get('beta.source-value', 90)
            gamma = entry.get('gamma.source-value', 90)

            atoms = []
            for n in range(len(cellpar[0])):
                try:
                    cell = Cell.fromcellpar([cellpar[0][n], cellpar[1][n], cellpar[2][n], alpha, beta, gamma])
                    atom = aseatoms(scaled_positions=basis, cell=cell, pbc=True)
                    if len(symbols) == len(atom.numbers):
                        atom.symbols = symbols
                    else:
                        atom.symbols = ['X' for _ in atom.numbers] if len(symbols) == 0 else symbols
                    atoms.append(atom)
                except Exception:
                    self.logger.error('Error generating structure.')
            return atoms

        def phonon_transform_function(archive, property_name, **kwargs):
            # Transform phonon frequencies to select data only for special kpoints
            data = np.array(archive.m_xpath(property_name))
            kpoints_special = kwargs.get('kpoints')
            if kpoints_special is None:
                cell = Cell(archive.run[-1].system[-1].atoms.lattice_vectors.to('angstrom').magnitude)
                kpoints_special = cell.get_bravais_lattice().get_special_points_array()

            kpoints = np.array(archive.m_xpath('%s.kpoints' % property_name.rsplit('.', 1)[0]))

            data_special = []
            for kpoint in kpoints_special:
                index = np.where((kpoints[:, :, 0] == kpoint[0]) & (kpoints[:, :, 1] == kpoint[1]) & (kpoints[:, :, 2] == kpoint[2]))
                if np.size(index) == 0:
                    continue
                data_special.append(data[index[0][0], :, index[0][1], :])

            return np.array(data_special).flatten()

        # first entry is the parser-generated header used identify an  open-kim
        for entry in self.entries:
            sec_run = self.archive.m_create(Run)
            sec_run.program = Program(name='OpenKIM', version=entry.get('meta.runner.short-id'))

            compile_date = entry.get('meta.created_on')
            if compile_date is not None:
                dt = datetime.strptime(compile_date, '%Y-%m-%d %H:%M:%S.%f') - datetime(1970, 1, 1)
                sec_run.program.compilation_datetime = dt.total_seconds()

            # openkim metadata
            sec_run.x_openkim_meta = {key: entry.pop(key) for key in list(entry.keys()) if key.startswith('meta.')}

            atoms = get_atoms(entry)
            for atom in atoms:
                sec_system = sec_run.m_create(System)
                sec_atoms = sec_system.m_create(Atoms)
                sec_atoms.labels = atom.get_chemical_symbols()
                sec_atoms.positions = atom.get_positions() * ureg.angstrom
                sec_atoms.lattice_vectors = atom.get_cell().array * ureg.angstrom
                sec_atoms.periodic = [True, True, True]
            try:
                self.material['space_group_number'] = Spacegroup(entry['space-group.source-value']).no
                self.material['structure_name'] = entry['short-name.source-value'][-1]
            except Exception:
                pass

            # model parameters
            model = sec_run.x_openkim_meta.get('meta.model')
            if model is not None:
                sec_method = sec_run.m_create(Method)
                sec_method.force_field = ForceField(model=[Model(
                    name=model,
                    reference='https://openkim.org/id/%s' % model)])

            energies = get_value(entry, 'cohesive-potential-energy.si-value', True)
            for n, energy in enumerate(energies):
                sec_scc = sec_run.m_create(Calculation)
                sec_scc.energy = Energy(total=EnergyEntry(value=energy))

            temperatures = get_value(entry, 'temperature.si-value', True)
            for n, temperature in enumerate(temperatures):
                sec_scc = sec_run.calculation[n] if sec_run.calculation else sec_run.m_create(Calculation)
                sec_scc.thermodynamics.append(Thermodynamics(temperature=temperature))

            stress = get_value(entry, 'cauchy-stress.si-value')
            if stress is not None:
                sec_scc = sec_run.calculation[-1] if sec_run.calculation else sec_run.m_create(Calculation)
                stress_tensor = np.zeros((3, 3))
                stress_tensor[0][0] = stress[0]
                stress_tensor[1][1] = stress[1]
                stress_tensor[2][2] = stress[2]
                stress_tensor[1][2] = stress_tensor[2][1] = stress[3]
                stress_tensor[0][2] = stress_tensor[2][0] = stress[4]
                stress_tensor[0][1] = stress_tensor[1][0] = stress[5]
                sec_scc.stress = Stress(total=StressEntry(value=stress_tensor))
                # TODO add nomad error, which property corresponding to stress

            for key, val in entry.items():
                key = 'x_openkim_%s' % re.sub(r'\W', '_', key)
                try:
                    setattr(sec_run, key, val)
                except Exception:
                    pass

            # workflow
            property_id = entry.get('property-id', '')
            # elastic constants
            if 'elastic-constants' in property_id:
                sec_workflow = self.archive.m_create(Workflow)
                sec_workflow.type = 'elastic'
                sec_elastic = sec_workflow.m_create(Elastic)
                cij = [[get_value(entry, f'c{i}{j}.si-value', default=0) for j in range(1, 7)] for i in range(1, 7)]
                sg = self.material.get('space_group_number', 0)
                if sg <= 74:
                    # triclinic / monoclinic / orthorhombic
                    pass
                elif sg <= 142:  # tetragonal
                    cij[1][1] = cij[0][0]
                    cij[1][2] = cij[0][2]
                    cij[1][5] = -cij[0][5]
                    cij[4][4] = cij[3][3]
                elif sg <= 167:  # trigonal
                    cij[1][1] = cij[0][0]
                    cij[1][2] = cij[0][2]
                    cij[1][3] = -cij[0][3]
                    cij[1][4] = -cij[0][4]
                    cij[3][5] = -cij[0][4]
                    cij[4][5] = cij[0][3]
                    cij[5][5] = (cij[0][0] - cij[0][1]) / 2
                elif sg <= 194:  # hexagonal
                    cij[1][1] = cij[0][0]
                    cij[1][2] = cij[0][2]
                    cij[4][4] = cij[3][3]
                    cij[5][5] = (cij[0][0] - cij[0][1]) / 2
                elif sg <= 230:  # cubic
                    cij[0][2] = cij[0][1]
                    cij[1][1] = cij[0][0]
                    cij[1][2] = cij[0][1]
                    cij[2][2] = cij[0][0]
                    cij[4][4] = cij[3][3]
                    cij[5][5] = cij[3][3]

                sec_elastic.elastic_constants_matrix_second_order = symmetrize_matrix(cij)

                try:
                    self.calculate_nomad_error('workflow[-1].elastic.elastic_constants_matrix_second_order')
                except Exception:
                    self.logger.error('Failed to calculate nomad error.')

                if 'strain-gradient' in property_id:
                    # TODO implement symmetry
                    dij = [[get_value(entry, f'd-{i}-{j}.si-value', default=0) for i in range(1, 19)] for j in range(1, 19)]
                    sec_elastic.elastic_constants_gradient_matrix_second_order = symmetrize_matrix(dij)

                if 'excess.si-value' in entry:
                    sec_elastic.x_openkim_excess = entry['excess.si-value']

            if 'gamma-surface' in property_id or 'stacking-fault' in property_id or 'twinning-fault' in property_id:
                sec_workflow = self.archive.m_create(Workflow)
                sec_workflow.type = 'interface'
                sec_interface = sec_workflow.m_create(Interface)
                if 'gamma-surface.si-value' in entry:
                    directions, displacements = [], []
                    for key in entry.keys():
                        direction = re.match(r'fault-plane-shift-fraction-(\d+).source-value', key)
                        if direction:
                            directions.append(direction.group(1))
                            displacements.append(entry[key])
                    sec_interface.dimensionality = len(directions)
                    sec_interface.shift_direction = directions
                    sec_interface.displacement_fraction = displacements
                    sec_interface.gamma_surface = entry['gamma-surface.si-value']

                if 'fault-plane-energy.si-value' in entry:
                    sec_interface.dimensionality = 1
                    sec_interface.displacement_fraction = [entry['fault-plane-shift-fraction.source-value']]
                    sec_interface.energy_fault_plane = entry['fault-plane-energy.si-value']

                sec_interface.energy_extrinsic_stacking_fault = entry.get('extrinsic-stacking-fault-energy.si-value')
                sec_interface.energy_intrinsic_stacking_fault = entry.get('intrinsic-stacking-fault-energy.si-value')
                sec_interface.energy_unstable_stacking_fault = entry.get('unstable-stacking-energy.si-value')
                sec_interface.energy_unstable_twinning_fault = entry.get('unstable-twinning-energy.si-value')
                sec_interface.slip_fraction = entry.get('unstable-slip-fraction.source-value')

            if 'phonon-dispersion' in property_id:
                sec_workflow = self.archive.m_create(Workflow)
                sec_workflow.type = 'phonon'
                sec_phonon = sec_workflow.m_create(Phonon)
                if 'response-frequency.si-value' in entry:
                    sec_scc = sec_run.calculation[-1] if sec_run.calculation else sec_run.m_create(Calculation)
                    sec_bandstructure = sec_scc.m_create(BandStructure, Calculation.band_structure_phonon)
                    # TODO find a way to segment the frequencies
                    sec_segment = sec_bandstructure.m_create(BandEnergies)
                    energies = entry['response-frequency.si-value']
                    if len(np.shape(energies)) == 1:
                        energies = energies[0]
                    sec_segment.energies = [energies]
                    try:
                        wavevector = entry['wave-vector-direction.si-value']
                        cell = sec_run.system[-1].atoms.lattice_vectors.magnitude
                        # TODO how about spin-polarized case, not sure about calculation of kpoints value
                        sec_segment.kpoints = np.dot(wavevector, cell)
                    except Exception:
                        pass

                try:
                    self.calculate_nomad_error('run[-1].calculation[-1].band_structure_phonon[-1].segment[*].energies', phonon_transform_function)
                except Exception:
                    self.logger.error('Failed to calculate nomad error.')

                if 'wave-number.si-value' in entry:
                    sec_phonon.x_openkim_wave_number = [entry['wave-number.si-value']]

        # write archive to file
        if filename is not None:
            with open(filename, 'w') as f:
                json.dump(self.archive.m_to_dict(), f, indent=4)


class OpenKIMParser:
    def __init__(self):
        pass

    def parse(self, filepath, archive, logger):
        logger = logger if logger is not None else logging.getLogger('__name__')

        try:
            with open(os.path.abspath(filepath), 'rt') as f:
                archive_data = json.load(f)
        except Exception:
            logger.error('Error reading openkim archive')
            return

        if isinstance(archive_data, dict) and archive_data.get('run') is not None:
            archive.m_update_from_dict(archive_data)
            return

        # support for old version
        if isinstance(archive_data, dict) and archive_data.get('QUERY') is not None:
            archive_data = archive_data['QUERY']

        converter = Converter(archive_data, logger)
        converter.archive = archive
        converter.convert()


def openkim_entries_to_nomad_archive(entries, filename=None):
    if isinstance(entries, str):
        if filename is None:
            filename = 'openkim_archive_%s.json' % os.path.basename(entries).rstrip('.json')
        with open(entries) as f:
            entries = json.load(f)

    Converter(entries).convert(filename)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np
from elasticsearch_dsl import Text

from ase.data import chemical_symbols
from ase.formula import Formula

from nomad import config
from nomad import atomutils
from nomad.datamodel.metainfo.measurements import Spectrum
from nomad.datamodel.metainfo.workflow import EquationOfState, EOSFit, Workflow
from nomad.metainfo.elasticsearch_extension import (
    Elasticsearch,
    material_type,
    material_entry_type,
    get_tokenizer
)

from nomad.metainfo import (
    MSection,
    Section,
    SubSection,
    Quantity,
    MEnum,
    Package,
    Datetime,
)

m_package = Package()

from nomad.datamodel.optimade import Species  # noqa
from nomad.datamodel.metainfo.simulation.calculation import (
    Dos,
    BandStructure as BandStructureCalculation,
    BandEnergies,
    DosValues,
    Calculation)  # noqa
from nomad.datamodel.metainfo.simulation.method import (
    BasisSet, Scf, Electronic, Smearing, GW as GWMethod
)  # noqa
from nomad.datamodel.metainfo.workflow import (
    GeometryOptimization as MGeometryOptimization, Thermodynamics
)  # noqa


unavailable = 'unavailable'
not_processed = 'not processed'
structure_classes = [
    'bulk',
    'surface',
    '2D',
    '1D',
    'molecule / cluster',
    'atom',
    unavailable,
    not_processed,
]
bravais_lattices = [
    'aP',
    'mP',
    'mS',
    'oP',
    'oS',
    'oF',
    'oI',
    'tP',
    'tI',
    'hP',
    'hR',
    'cP',
    'cF',
    'cI',
]
crystal_systems = [
    'triclinic',
    'monoclinic',
    'orthorhombic',
    'tetragonal',
    'trigonal',
    'hexagonal',
    'cubic',
]
xc_treatments = {
    'gga': 'GGA',
    'hf_': 'HF',
    'oep': 'OEP',
    'hyb': 'hybrid',
    'mgg': 'meta-GGA',
    'vdw': 'vdW',
    'lda': 'LDA',
}
basis_set_types = [
    '(L)APW+lo',
    'gaussians',
    'numeric AOs',
    'plane waves',
    'psinc functions',
    'real-space grid',
    unavailable,
    not_processed,
]
core_electron_treatments = [
    'full all electron',
    'all electron frozen core',
    'pseudopotential',
    unavailable,
]


def variants_formula(value):
    ''' Creates several common variants for the given formula.'''
    formula = Formula(value)
    formats = ['hill', 'metal', 'abc']
    formulas = [value] + [formula.format(f) for f in formats]
    return list(set(formulas))


tokenizer_formula = get_tokenizer(r'[A-Z][a-z]?\d*')


class BandGap(MSection):
    m_def = Section(
        description='''
        Band gap information for each spin channel.
        '''
    )
    index = Quantity(
        type=np.dtype(np.int32),
        description='''
        Spin channel index.
        ''',
        a_elasticsearch=Elasticsearch(material_entry_type),
    )
    value = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Band gap value. Value of zero corresponds to not having a band gap.
        ''',
        a_elasticsearch=Elasticsearch(material_entry_type),
    )
    type = Quantity(
        type=MEnum('direct', 'indirect'),
        shape=[],
        description='''
        Band gap type.
        ''',
        a_elasticsearch=[
            Elasticsearch(material_entry_type),
            Elasticsearch(suggestion='simple')
        ],
    )
    energy_highest_occupied = Quantity(
        type=np.dtype(np.float64),
        unit='joule',
        shape=[],
        description='''
        The highest occupied energy.
        ''',
    )
    energy_lowest_unoccupied = Quantity(
        type=np.dtype(np.float64),
        unit='joule',
        shape=[],
        description='''
        The lowest unoccupied energy.
        ''',
    )


class WyckoffSet(MSection):
    m_def = Section(
        description='''
        Section for storing Wyckoff set information. Only available for
        conventional cells that have undergone symmetry analysis.
        '''
    )
    wyckoff_letter = Quantity(
        type=str,
        description='''
        The Wyckoff letter for this set.
        '''
    )
    indices = Quantity(
        type=np.dtype('i4'),
        shape=['1..*'],
        description='''
        Indices of the atoms belonging to this group.
        '''
    )
    element = Quantity(
        type=str,
        description='''
        Chemical element at this Wyckoff position.
        '''
    )
    x = Quantity(
        type=np.dtype(np.float64),
        description='''
        The free parameter x if present.
        '''
    )
    y = Quantity(
        type=np.dtype(np.float64),
        description='''
        The free parameter y if present.
        '''
    )
    z = Quantity(
        type=np.dtype(np.float64),
        description='''
        The free parameter z if present.
        '''
    )


class LatticeParameters(MSection):
    m_def = Section(
        description='''
        Lattice parameters of a cell.
        ''',
    )
    a = Quantity(
        type=np.dtype(np.float64),
        unit='m',
        description='''
        Length of the first basis vector.
        ''',
        a_elasticsearch=Elasticsearch(material_entry_type),
    )
    b = Quantity(
        type=np.dtype(np.float64),
        unit='m',
        description='''
        Length of the second basis vector.
        ''',
        a_elasticsearch=Elasticsearch(material_entry_type),
    )
    c = Quantity(
        type=np.dtype(np.float64),
        unit='m',
        description='''
        Length of the third basis vector.
        ''',
        a_elasticsearch=Elasticsearch(material_entry_type),
    )
    alpha = Quantity(
        type=np.dtype(np.float64),
        unit='radian',
        description='''
        Angle between second and third basis vector.
        ''',
        a_elasticsearch=Elasticsearch(material_entry_type),
    )
    beta = Quantity(
        type=np.dtype(np.float64),
        unit='radian',
        description='''
        Angle between first and third basis vector.
        ''',
        a_elasticsearch=Elasticsearch(material_entry_type),
    )
    gamma = Quantity(
        type=np.dtype(np.float64),
        unit='radian',
        description='''
        Angle between first and second basis vector.
        ''',
        a_elasticsearch=Elasticsearch(material_entry_type),
    )


class Structure(MSection):
    m_def = Section(
        description='''
        Describes an atomistic structure.
        '''
    )
    dimension_types = Quantity(
        type=int,
        shape=[3],
        default=[0, 0, 0],
        description='''
        List of three integers. For each of the three directions indicated by
        the three lattice vectors (see property lattice_vectors). This list
        indicates if the direction is periodic (value 1) or non-periodic (value
        0). Note: the elements in this list each refer to the direction of the
        corresponding entry in lattice_vectors and not the Cartesian x, y, z
        directions.
        '''
    )
    nperiodic_dimensions = Quantity(
        type=int,
        derived=lambda a: sum(a.dimension_types),
        description='''
        An integer specifying the number of periodic dimensions in the
        structure, equivalent to the number of non-zero entries in
        dimension_types.
        ''',
        a_elasticsearch=Elasticsearch(material_entry_type),
    )
    lattice_vectors = Quantity(
        type=np.dtype('float64'),
        shape=[3, 3],
        unit='m',
        description='''
        The three lattice vectors in Cartesian coordinates.
        '''
    )
    cartesian_site_positions = Quantity(
        type=np.dtype('float64'),
        shape=['n_sites', 3],
        unit='m',
        description='''
        Cartesian positions of each site. A site is an atom, a site potentially
        occupied by an atom, or a placeholder for a virtual mixture of atoms
        (e.g., in a virtual crystal approximation).
        '''
    )
    n_sites = Quantity(
        type=int,
        default=0,
        derived=lambda a: len(a.cartesian_site_positions) if a.cartesian_site_positions is not None else 0,
        description='''
        An integer specifying the length of the cartesian_site_positions property.
        ''',
        a_elasticsearch=Elasticsearch(material_entry_type),
    )
    species_at_sites = Quantity(
        type=str,
        shape=['n_sites'],
        description='''
        Name of the species at each site (where values for sites are specified with the same
        order of the cartesian_site_positions property). The properties of the species are
        found in the species property.
        '''
    )
    cell_volume = Quantity(
        type=np.dtype(np.float64),
        unit='m ** 3',
        description='''
        Volume of the cell.
        ''',
        a_elasticsearch=Elasticsearch(material_entry_type),
    )
    atomic_density = Quantity(
        type=np.dtype(np.float64),
        unit='1 / m ** 3',
        description='''
        Atomic density of the material (atoms/volume).'
        '''
    )
    mass_density = Quantity(
        type=np.dtype(np.float64),
        unit='kg / m ** 3',
        description='''
        Mass density of the material.
        '''
    )
    species = SubSection(sub_section=Species.m_def, repeats=True)
    lattice_parameters = SubSection(sub_section=LatticeParameters.m_def)


class StructureOriginal(Structure):
    m_def = Section(
        description='''
        Contains a selected representative structure from the the original
        data.
        '''
    )


class StructurePrimitive(Structure):
    m_def = Section(
        description='''
        Contains the primitive structure that is derived from
        structure_original. This primitive stucture has been idealized and the
        conventions employed by spglib are used.
        '''
    )


class StructureConventional(Structure):
    m_def = Section(
        description='''
        Contains the conventional structure that is derived from
        structure_original. This conventional stucture has been idealized and
        the conventions employed by spglib are used.
        '''
    )
    wyckoff_sets = SubSection(sub_section=WyckoffSet.m_def, repeats=True)


class StructureOptimized(Structure):
    m_def = Section(
        description='''
        Contains a structure that is the result of a geometry optimization.
        '''
    )


class Structures(MSection):
    m_def = Section(
        description='''
        Contains full atomistic representations of the material in different
        forms.
        ''',
    )
    structure_original = SubSection(
        sub_section=StructureOriginal.m_def,
        repeats=False,
    )
    structure_conventional = SubSection(
        sub_section=StructureConventional.m_def,
        repeats=False,
    )
    structure_primitive = SubSection(
        sub_section=StructurePrimitive.m_def,
        repeats=False,
    )


class Symmetry(MSection):
    m_def = Section(
        description='''
        Section containing information about the symmetry of the material. All
        of these properties are derived by running a symmetry analysis on a
        representative geometry from the original data. This original geometry
        is stored in results.properties together with the primitive and
        conventional structures.
        '''
    )
    bravais_lattice = Quantity(
        type=MEnum(bravais_lattices),
        shape=[],
        description='''
        Identifier for the Bravais lattice in Pearson notation. The first lowercase letter
        identifies the crystal family and can be one of the following: a (triclinic), b
        (monoclinic), o (orthorhombic), t (tetragonal), h (hexagonal) or c (cubic). The
        second uppercase letter identifies the centring and can be one of the following: P
        (primitive), S (face centred), I (body centred), R (rhombohedral centring) or F
        (all faces centred).
        ''',
        a_elasticsearch=[
            Elasticsearch(material_type),
            Elasticsearch(suggestion='simple')
        ],
    )
    crystal_system = Quantity(
        type=MEnum(crystal_systems),
        shape=[],
        description='''
        Name of the crystal system.
        ''',
        a_elasticsearch=[
            Elasticsearch(material_type),
            Elasticsearch(suggestion='simple')
        ],
    )
    hall_number = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        The Hall number for this system.
        ''',
        a_elasticsearch=Elasticsearch(material_type),
    )
    hall_symbol = Quantity(
        type=str,
        shape=[],
        description='''
        The Hall symbol for this system.
        ''',
        a_elasticsearch=[
            Elasticsearch(material_type),
            Elasticsearch(suggestion='simple')
        ],
    )
    point_group = Quantity(
        type=str,
        shape=[],
        description='''
        Symbol of the crystallographic point group in the Hermann-Mauguin notation.
        ''',
        a_elasticsearch=[
            Elasticsearch(material_type),
            Elasticsearch(suggestion='simple')
        ],
    )
    space_group_number = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Specifies the International Union of Crystallography (IUC) number of the 3D space
        group of this system.
        ''',
        a_elasticsearch=Elasticsearch(material_type),
    )
    space_group_symbol = Quantity(
        type=str,
        shape=[],
        description='''
        The International Union of Crystallography (IUC) short symbol of the 3D
        space group of this system.
        ''',
        a_elasticsearch=[
            Elasticsearch(material_type),
            Elasticsearch(suggestion='simple')
        ],
    )
    prototype_formula = Quantity(
        type=str,
        description='''
        The formula of the prototypical material for this structure.
        ''',
        a_elasticsearch=Elasticsearch(material_type),
    )
    prototype_aflow_id = Quantity(
        type=str,
        description='''
        The identifier of this structure in the AFLOW encyclopedia of
        crystallographic prototypes:
        http://www.aflowlib.org/prototype-encyclopedia/index.html
        ''',
        a_elasticsearch=[
            Elasticsearch(material_type),
            Elasticsearch(suggestion='simple')
        ],
    )
    structure_name = Quantity(
        type=str,
        description='''
        A common name for this structure, e.g. fcc, bcc.
        ''',
        a_elasticsearch=[
            Elasticsearch(material_type),
            Elasticsearch(suggestion='default')
        ],
    )
    strukturbericht_designation = Quantity(
        type=str,
        description='''
        Classification of the material according to the historically grown
        'strukturbericht'.
        ''',
        a_elasticsearch=[
            Elasticsearch(material_type),
            Elasticsearch(suggestion='simple')
        ],
    )


class Material(MSection):
    m_def = Section(
        description='''
        Contains information that is specific to bulk crystalline materials.
        '''
    )
    material_id = Quantity(
        type=str,
        description='''
        A fixed length, unique material identifier in the form of a hash
        digest.
        ''',
        a_elasticsearch=Elasticsearch(material_type, metrics=dict(n_materials='cardinality'))
    )
    material_name = Quantity(
        type=str,
        description='''
        Meaningful names for this a material if any can be assigned.
        ''',
        a_elasticsearch=[
            Elasticsearch(material_type),
            Elasticsearch(suggestion='default')
        ],
    )
    structural_type = Quantity(
        type=MEnum(structure_classes), default='not processed',
        description='''
        Classification based on structural features.
        ''',
        a_elasticsearch=[
            Elasticsearch(material_type),
            Elasticsearch(suggestion='default')
        ],
    )
    functional_type = Quantity(
        type=str,
        shape=['0..*'],
        description='''
        Classification based on the functional properties.
        ''',
        a_elasticsearch=[
            Elasticsearch(material_type, default_aggregation_size=20),
            Elasticsearch(suggestion='default')
        ],
    )
    compound_type = Quantity(
        type=str,
        shape=['0..*'],
        description='''
        Classification based on the chemical formula.
        ''',
        a_elasticsearch=[
            Elasticsearch(material_type, default_aggregation_size=20),
            Elasticsearch(suggestion='default')
        ],
    )
    elements = Quantity(
        type=MEnum(chemical_symbols),
        shape=['0..*'],
        default=[],
        description='''
        Names of the different elements present in the structure.
        ''',
        a_elasticsearch=[
            Elasticsearch(material_type, many_all=True),
            Elasticsearch(suggestion='simple')
        ]
    )
    n_elements = Quantity(
        type=int,
        default=0,
        derived=lambda s: len(s.elements),
        description='''
        Number of different elements in the structure as an integer.
        ''',
        a_elasticsearch=Elasticsearch(material_type),
    )
    elements_exclusive = Quantity(
        type=str,
        derived=lambda s: ' '.join(sorted(s.elements)),
        description='''
        String containing the chemical elements in alphabetical order and
        separated by a single whitespace. This quantity can be used for
        exclusive element searches where you want to find entries/materials
        with only certain given elements.
        ''',
        a_elasticsearch=Elasticsearch(material_type)
    )
    chemical_formula_descriptive = Quantity(
        type=str,
        description='''
            The chemical formula for a structure as a string in a form chosen by the API
            implementation.
        ''',
        a_elasticsearch=[
            Elasticsearch(material_type),
            Elasticsearch(suggestion=tokenizer_formula)
        ],
    )
    chemical_formula_reduced = Quantity(
        type=str,
        description='''
            The reduced chemical formula for a structure as a string with element symbols and
            integer chemical proportion numbers. The proportion number MUST be omitted if it is 1.
        ''',
        a_elasticsearch=[
            Elasticsearch(material_type),
            Elasticsearch(suggestion=tokenizer_formula)
        ],
    )
    chemical_formula_hill = Quantity(
        type=str,
        description='''
            The chemical formula for a structure in Hill form with element symbols followed by
            integer chemical proportion numbers. The proportion number MUST be omitted if it is 1.
        ''',
        a_elasticsearch=[
            Elasticsearch(material_type, normalizer=atomutils.get_formula_hill),
            Elasticsearch(suggestion=tokenizer_formula, variants=variants_formula)
        ],
    )
    chemical_formula_anonymous = Quantity(
        type=str,
        description='''
            The anonymous formula is the chemical_formula_reduced, but where the elements are
            instead first ordered by their chemical proportion number, and then, in order left to
            right, replaced by anonymous symbols A, B, C, ..., Z, Aa, Ba, ..., Za, Ab, Bb, ... and
            so on.
        ''',
        a_elasticsearch=[
            Elasticsearch(material_type),
            Elasticsearch(suggestion=tokenizer_formula)
        ],
    )
    chemical_formula_reduced_fragments = Quantity(
        type=str,
        shape=['*'],
        description='''
        The reduced formula separated into individual terms containing both the atom
        type and count. Used for searching parts of a formula.
        ''',
        a_elasticsearch=Elasticsearch(material_type, mapping=Text(multi=True)),
    )
    symmetry = SubSection(sub_section=Symmetry.m_def, repeats=False)


class DFT(MSection):
    m_def = Section(
        description='''
        Methodology for a DFT calculation.
        '''
    )
    basis_set_type = Quantity(
        type=MEnum(basis_set_types),
        default=unavailable,
        description='The used basis set functions.',
        a_elasticsearch=[
            Elasticsearch(material_entry_type),
            Elasticsearch(suggestion='default')
        ],
    )
    basis_set_name = BasisSet.name.m_copy()
    basis_set_name.m_annotations['elasticsearch'] = [
        Elasticsearch(material_entry_type),
        Elasticsearch(suggestion='default')
    ]
    core_electron_treatment = Quantity(
        type=MEnum(core_electron_treatments),
        default=unavailable,
        description='''
        How the core electrons are described.
        ''',
        a_elasticsearch=[
            Elasticsearch(material_entry_type),
            Elasticsearch(suggestion='default')
        ],
    )
    spin_polarized = Quantity(
        type=bool,
        description='''
        Whether the calculation is spin-polarized.
        ''',
        a_elasticsearch=Elasticsearch(material_entry_type),
    )
    scf_threshold_energy_change = Scf.threshold_energy_change.m_copy()
    scf_threshold_energy_change.m_annotations['elasticsearch'] = Elasticsearch(material_entry_type)
    van_der_Waals_method = Electronic.van_der_waals_method.m_copy()
    van_der_Waals_method.description = 'The used van der Waals method.'
    van_der_Waals_method.m_annotations['elasticsearch'] = [
        Elasticsearch(material_entry_type),
        Elasticsearch(suggestion='default')
    ]

    relativity_method = Electronic.relativity_method.m_copy()
    relativity_method.m_annotations['elasticsearch'] = [
        Elasticsearch(material_entry_type),
        Elasticsearch(suggestion='default')
    ]

    smearing_kind = Smearing.kind.m_copy()
    smearing_kind.m_annotations['elasticsearch'] = [
        Elasticsearch(material_entry_type),
        Elasticsearch(suggestion='default')
    ]

    smearing_width = Smearing.width.m_copy()
    smearing_width.m_annotations['elasticsearch'] = Elasticsearch(material_entry_type)
    xc_functional_type = Quantity(
        type=MEnum(list(xc_treatments.values()) + [unavailable, not_processed]),
        default=not_processed,
        description='The libXC based xc functional classification used in the simulation.',
        a_elasticsearch=Elasticsearch(material_entry_type, default_aggregation_size=100)
    )
    xc_functional_names = Quantity(
        type=str,
        default=[],
        shape=['*'],
        description='The list of libXC functional names that where used in this entry.',
        a_elasticsearch=[
            Elasticsearch(material_entry_type),
            Elasticsearch(suggestion='default')
        ]
    )


class GW(MSection):
    m_def = Section(
        description='''
        Methodology for a GW calculation.
        '''
    )
    type = GWMethod.type.m_copy()
    type.m_annotations['elasticsearch'] = [
        Elasticsearch(material_entry_type),
        Elasticsearch(suggestion='default')
    ]
    starting_point = Quantity(
        type=str,
        default=[],
        shape=['*'],
        description='The list of libXC functional names that were used for the ground state calculation.',
        a_elasticsearch=Elasticsearch(material_entry_type),
    )


class QuantumCircuit(MSection):
    processors = Quantity(type=str, shape=['0..*'])
    number_of_registers = Quantity(type=int)
    simulated = Quantity(type=bool)


class QuantumCMS(MSection):
    transformation = Quantity(type=str)
    quantum_computer_system = Quantity(type=str)
    quantum_computing_libraries = Quantity(type=str, shape=['0..*'])
    computation_datetime = Quantity(type=Datetime)
    number_of_shots = Quantity(type=int)
    quantum_volume = Quantity(type=int)
    quantum_circuit = SubSection(sub_section=QuantumCircuit)


class Simulation(MSection):
    m_def = Section(
        description='''
        Contains method details for a simulation entry.
        '''
    )
    program_name = Quantity(
        type=str,
        default='not processed',
        description='The name of the used program.',
        a_elasticsearch=[
            Elasticsearch(material_entry_type),
            Elasticsearch(suggestion='default')
        ],
    )
    program_version = Quantity(
        type=str,
        default='not processed',
        description='The version of the used program.',
        a_elasticsearch=[
            Elasticsearch(material_entry_type),
            Elasticsearch(suggestion='default')
        ],
    )
    dft = SubSection(sub_section=DFT.m_def, repeats=False)
    gw = SubSection(sub_section=GW.m_def, repeats=False)
    quantum_cms = SubSection(sub_section=QuantumCMS.m_def, repeats=False)


class Method(MSection):
    m_def = Section(
        description='''
        Contains a summary of the methodology that has been used in this entry.
        This methodology applies to all of the reported properties and
        determines the result of a single energy evalution. The individual
        properties may be further methodological details affect e.g. the
        sampling.
        '''
    )
    method_id = Quantity(
        type=str,
        description='''
        Identifier for the used method. Only available for a subset of entries
        for which the methodology has been identified with precision.
        ''',
        a_elasticsearch=Elasticsearch(material_entry_type),
    )
    equation_of_state_id = Quantity(
        type=str,
        description='''
        Identifier that can be used to group entries within an equation of
        state calculation. Only available for a subset of entries for which the
        structure and methodology have been identified with precision.
        ''',
    )
    parameter_variation_id = Quantity(
        type=str,
        description='''
        Identifier that can be used to group entries that target the same
        structure but with varying parameter settings. Only available for a
        subset of entries for which the structure and methodology have been
        identified with precision.
        ''',
    )
    method_name = Quantity(
        type=MEnum(['DFT', 'GW', 'EELS', 'XPS', config.services.unavailable_value]),
        description='''
        Common name for the used method.
        ''',
        a_elasticsearch=[
            Elasticsearch(material_entry_type),
            Elasticsearch(suggestion='default')
        ],
    )

    workflow_name = Workflow.type.m_copy()
    workflow_name.shape = ['*']
    workflow_name.m_annotations['elasticsearch'] = [
        Elasticsearch(material_entry_type),
        Elasticsearch(suggestion='default')
    ]
    simulation = SubSection(sub_section=Simulation.m_def, repeats=False)


class DOS(MSection):
    m_def = Section(
        description='''
        Base class for density of states information.
        ''',
    )
    energies = Quantity(
        type=Dos.energies,
        description='''
        Array containing the set of discrete energy values for the density of
        states (DOS).
        ''',
    )
    total = Quantity(
        type=DosValues,
        shape=['*'],
        description='''
        Density of states (DOS) values normalized with unit cell volume and
        number of atoms.
        ''',
    )


class DOSPhonon(DOS):
    m_def = Section(
        description='''
        Contains the total phonon density of states.
        ''',
    )


class DOSElectronic(DOS):
    m_def = Section(
        description='''
        Contains the total electronic density of states.
        ''',
    )
    spin_polarized = Quantity(
        type=bool,
        description='''
        Whether the DOS is spin-polarized, i.e. is contains channels for both
        spin values.
        ''',
        a_elasticsearch=Elasticsearch(material_entry_type),
    )
    band_gap = SubSection(
        sub_section=BandGap.m_def,
        repeats=True,
        a_elasticsearch=Elasticsearch(material_entry_type, nested=True)
    )
    energy_fermi = Quantity(
        type=np.dtype(np.float64),
        unit='joule',
        shape=[],
        description='''
        Fermi energy.
        '''
    )


class BandStructure(MSection):
    m_def = Section(
        description='''
        Base class for band structure information.
        ''',
    )
    reciprocal_cell = Quantity(
        type=BandStructureCalculation.reciprocal_cell,
        description='''
        The reciprocal cell within which the band structure is calculated.
        ''',
    )
    segment = Quantity(
        type=BandEnergies,
        shape=['*'],
        description='''
        Collection of linear path segments in the reciprocal space. The
        segments are represented as third-order tensors: one dimension for the
        spin channels, one for the sequence of reciprocal space points for the
        segment, and one for the sequence of eigenvalues at a given point.
        ''',
    )
    path_standard = Quantity(
        type=str,
        shape=[],
        description='''
        String that identifies the possible standard used in sampling the
        reciprocal space.
        ''',
    )


class BandStructurePhonon(BandStructure):
    m_def = Section(
        description='''
        This section stores information on a vibrational band structure
        evaluation along one-dimensional pathways in the reciprocal space.
        '''
    )


class BandStructureElectronic(BandStructure):
    m_def = Section(
        description='''
        This section stores information on a electonic band structure
        evaluation along one-dimensional pathways in the reciprocal space.
        '''
    )
    spin_polarized = Quantity(
        type=bool,
        description='''
        Whether the band structure is spin-polarized, i.e. is contains channels
        for both spin values.
        ''',
        a_elasticsearch=Elasticsearch(material_entry_type),
    )
    band_gap = SubSection(
        sub_section=BandGap.m_def,
        repeats=True,
        a_elasticsearch=Elasticsearch(material_entry_type, nested=True)
    )
    energy_fermi = Quantity(
        type=np.dtype(np.float64),
        unit='joule',
        shape=[],
        description='''
        Fermi energy.
        '''
    )


class HeatCapacityConstantVolume(MSection):
    m_def = Section(
        description='''
        Contains the values of the specific (per mass) and isochoric (constant
        volume) heat capacity at different temperatures.
        '''
    )
    heat_capacities = Quantity(
        type=Thermodynamics.heat_capacity_c_v,
        shape=[],
        description='''
        Specific heat capacity values at constant volume.
        ''',
    )
    temperatures = Quantity(
        type=Thermodynamics.temperature,
        description='''
        The temperatures at which heat capacities are calculated.
        ''',
    )


class EnergyFreeHelmholtz(MSection):
    m_def = Section(
        description='''
        Contains the values of the Helmholtz free energy per atom at constant
        volume and at different temperatures.
        '''
    )
    energies = Quantity(
        type=Thermodynamics.vibrational_free_energy_at_constant_volume,
        shape=[],
        description='''
        The Helmholtz free energies per atom at constant volume.
        ''',
    )
    temperatures = Quantity(
        type=Thermodynamics.temperature,
        description='''
        The temperatures at which Helmholtz free energies are calculated.
        ''',
    )


class VibrationalProperties(MSection):
    m_def = Section(
        description='''
        Vibrational properties.
        ''',
    )
    band_structure_phonon = SubSection(sub_section=BandStructurePhonon.m_def, repeats=False)
    dos_phonon = SubSection(sub_section=DOSPhonon.m_def, repeats=False)
    heat_capacity_constant_volume = SubSection(sub_section=HeatCapacityConstantVolume.m_def, repeats=False)
    energy_free_helmholtz = SubSection(sub_section=EnergyFreeHelmholtz.m_def, repeats=False)


class EnergyVolumeCurve(MSection):
    m_def = Section(
        description='''
        Energy volume curve.
        ''',
    )
    type = Quantity(
        type=MEnum(
            'raw',
            'mie_gruneisen',
            'pack_evans_james',
            'vinet',
            'tait',
            'birch_euler',
            'pourier_tarantola',
            'birch_lagrange',
            'murnaghan',
        ),
        a_elasticsearch=[
            Elasticsearch(material_entry_type),
            Elasticsearch(suggestion='default')
        ],
    )
    volumes = Quantity(type=EquationOfState.volumes)
    energies_raw = Quantity(type=EquationOfState.energies)
    energies_fit = Quantity(type=EOSFit.fitted_energies)


class BulkModulus(MSection):
    m_def = Section(
        description='''
        Contains bulk modulus values calculated with different methodologies.
        '''
    )
    type = Quantity(
        type=MEnum(
            'mie_gruneisen',
            'pack_evans_james',
            'vinet',
            'tait',
            'birch_euler',
            'pourier_tarantola',
            'birch_lagrange',
            'murnaghan',
            'voigt_average',
            'reuss_average',
            'voigt_reuss_hill_average',
        ),
        description='Describes the methodology for obtaining the value.',
        a_elasticsearch=[
            Elasticsearch(material_entry_type),
            Elasticsearch(suggestion='default')
        ],
    )
    value = Quantity(
        type=np.dtype(np.float64),
        description='Bulk modulus value.',
        unit='pascal',
        a_elasticsearch=Elasticsearch(material_entry_type),
    )


class ShearModulus(MSection):
    m_def = Section(
        description='''
        Contains shear modulus values calculated with different methodologies.
        ''',
    )
    type = Quantity(
        type=MEnum(
            'voigt_average',
            'reuss_average',
            'voigt_reuss_hill_average',
        ),
        description='Describes the methodology for obtaining the value.',
        a_elasticsearch=[
            Elasticsearch(material_entry_type),
            Elasticsearch(suggestion='default')
        ],
    )
    value = Quantity(
        type=np.dtype(np.float64),
        description='Shear modulus value.',
        unit='pascal',
        a_elasticsearch=Elasticsearch(material_entry_type),
    )


class GeometryOptimization(MSection):
    m_def = Section(
        description='''
        Geometry optimization results and settings.
        ''',
    )
    trajectory = Quantity(
        type=Calculation,
        shape=['0..*'],
        description='''
        List of references to each section_single_configuration_calculation in
        the optimization trajectory.
        ''',
    )
    energies = Quantity(
        type=MGeometryOptimization.energies,
        description='''
        List of energy_total values gathered from the single configuration
        calculations that are a part of the optimization trajectory.
        ''',
    )
    structure_optimized = SubSection(
        sub_section=StructureOptimized.m_def,
        repeats=False,
    )
    type = MGeometryOptimization.type.m_copy()
    convergence_tolerance_energy_difference = MGeometryOptimization.convergence_tolerance_energy_difference.m_copy()
    convergence_tolerance_energy_difference.m_annotations['elasticsearch'] = Elasticsearch(material_entry_type)
    convergence_tolerance_force_maximum = MGeometryOptimization.convergence_tolerance_force_maximum.m_copy()
    convergence_tolerance_force_maximum.m_annotations['elasticsearch'] = Elasticsearch(material_entry_type)
    final_force_maximum = MGeometryOptimization.final_force_maximum.m_copy()
    final_force_maximum.m_annotations['elasticsearch'] = Elasticsearch(material_entry_type)
    final_energy_difference = MGeometryOptimization.final_energy_difference.m_copy()
    final_energy_difference.m_annotations['elasticsearch'] = Elasticsearch(material_entry_type)
    final_displacement_maximum = MGeometryOptimization.final_displacement_maximum.m_copy()
    final_displacement_maximum.m_annotations['elasticsearch'] = Elasticsearch(material_entry_type)


class MechanicalProperties(MSection):
    m_def = Section(
        description='''
        Mechanical properties.
        ''',
    )
    energy_volume_curve = SubSection(sub_section=EnergyVolumeCurve.m_def, repeats=True)
    bulk_modulus = SubSection(
        sub_section=BulkModulus.m_def,
        repeats=True,
        a_elasticsearch=Elasticsearch(material_entry_type, nested=True)
    )
    shear_modulus = SubSection(
        sub_section=ShearModulus.m_def,
        repeats=True,
        a_elasticsearch=Elasticsearch(material_entry_type, nested=True)
    )


class ElectronicProperties(MSection):
    m_def = Section(
        description='''
        Electronic properties.
        ''',
    )
    band_structure_electronic = SubSection(sub_section=BandStructureElectronic.m_def, repeats=False)
    dos_electronic = SubSection(sub_section=DOSElectronic.m_def, repeats=False)


class SpectroscopyProperties(MSection):
    m_def = Section(
        description='''
        Spectroscopic properties.
        ''',
    )
    spectrum = Quantity(type=Spectrum)


class Properties(MSection):
    m_def = Section(
        description='''
        Contains the physical properties that have been calculated or used in
        this entry.
        '''
    )
    structures = SubSection(sub_section=Structures.m_def, repeats=False)
    vibrational = SubSection(sub_section=VibrationalProperties.m_def, repeats=False)
    electronic = SubSection(sub_section=ElectronicProperties.m_def, repeats=False)
    mechanical = SubSection(sub_section=MechanicalProperties.m_def, repeats=False)
    spectroscopy = SubSection(sub_section=SpectroscopyProperties.m_def, repeats=False)
    geometry_optimization = SubSection(sub_section=GeometryOptimization.m_def, repeats=False)

    n_calculations = Quantity(
        type=int,
        description='''
        The number of performed single configuration calculations.'
        ''',
        a_elasticsearch=Elasticsearch(material_entry_type, metrics=dict(n_calculations='sum')),
    )
    available_properties = Quantity(
        type=str,
        shape=['0..*'],
        description='Subset of the property names that are present in this entry.',
        a_elasticsearch=Elasticsearch(material_entry_type),
    )


class Results(MSection):
    m_def = Section(
        description='''
        Contains a summary of the entry contents.
        '''
    )
    material = SubSection(sub_section=Material.m_def, repeats=False)
    method = SubSection(sub_section=Method.m_def, repeats=False)
    properties = SubSection(sub_section=Properties.m_def, repeats=False)


m_package.__init_metainfo__()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
The NOMAD meta-info allows to define schemas for physics data independent of the used
storage format. It allows to define physics quantities with types, complex shapes
(vetors, matrices, etc.), units, links, and descriptions. It allows to organize large
amounts of these quantities in containment hierarchies of extendable sections, references
between sections, and additional quantity categories.

NOMAD uses the meta-info to define all archive data, repository meta-data, (and encyclopedia
data). The meta-info provides a convenient Python interface to create,
manipulate, and access data. We also use it to map data to various storage formats,
including JSON, (HDF5), mongodb, and elastic search.
'''


from .metainfo import (
    MSectionBound,
    MSection,
    MCategory,
    Definition,
    Property,
    Quantity,
    SubSection,
    Section,
    Category,
    Package,
    Environment,
    MEnum,
    Datetime,
    Capitalized,
    MProxy,
    MetainfoError,
    DeriveError,
    MetainfoReferenceError,
    DataType,
    Reference,
    SectionReference,
    QuantityReference,
    File,
    Datetime,
    Unit,
    JSON,
    Dimension,
    Bytes,
    Context,
    m_package,
    Annotation,
    DefinitionAnnotation,
    SectionAnnotation,
    SectionProxy,
    derived,
    constraint,
    units)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
Configuration for JuypterHUB. We use JuypterHUB to run remote tools, like jupyter, on
NOMAD servers.
'''

from dockerspawner.dockerspawner import DockerSpawner
from jupyterhub.handlers.base import BaseHandler
from oauthenticator.generic import GenericOAuthenticator
import os
import os.path
import requests
import json

from nomad import config, infrastructure

# TODO The AnonymousLogin kinda works, but it won't logout or remove old containers,
# if the user is still logged in non anonymously.


class AnonymousLoginHandler(BaseHandler):
    async def get(self):
        # TODO somehow set a fake user to avoid login
        # - read cookie and use cookie, or create, set, and use cookie
        await self.login_user(data='anonymous')
        self.redirect(self.get_next_url())


class NomadAuthenticator(GenericOAuthenticator):
    '''
    A custom OAuth authenticator. It can be used with NOMAD's keycloak. It imeplement
    a `pre_spawn_start` hook that passes user upload mounts to the spawner.
    '''

    def get_handlers(self, app):
        ''' Add an additional handler for anonymous logins. '''
        handlers = super().get_handlers(app)
        return handlers + [(r'/anonymous_login', AnonymousLoginHandler)]

    async def authenticate(self, handler, data=None):
        if data == 'anonymous':
            return 'anonymous'

        try:
            return await super().authenticate(handler, data=data)
        except Exception as e:
            # The default authenticate has failed, e.g. due to missing credentials.
            # Check if there is bearer authorization and try to identify the user with
            # this. Otherwise, propagate the exception.
            authorization = handler.request.headers.get('Authorization', None)
            if not (authorization and authorization.lower().startswith('bearer ')):
                raise e

        # Use the access token to authenticate the user
        access_token = authorization[7:]
        payload = infrastructure.keycloak.decode_access_token(access_token)

        authenticated = {
            'name': payload['preferred_username'],
            'auth_state': {
                'access_token': access_token,
                'refresh_token': None,  # It is unclear how we can get the refresh
                                        # token. The only way is through a proper oauth
                                        # flow. But we can't do that if the hub is only
                                        # used via API and not the hub pages.
                                        # For now the assumption is that
                                        # - we implicitly "refresh" the access token with
                                        #   each API call that causes this to run
                                        # - we assume that jhub we automatically go through
                                        #   a oauth flow, once the access token is expired
                                        #   and it does not find a stored refresh token.
                'oauth_user': payload,
                'scope': []}}
        # This will save the users authstate
        await handler.auth_to_user(authenticated)

        return authenticated

    async def pre_spawn_start(self, user, spawner):
        '''
        Uses the user credentials to request all staging uploads and pass the
        respective path as volume host mounts to the spawner.
        '''
        # This is guacamole specific
        # linuxserver/webtop guacamole-lite based guacamole client use SUBFOLDER to
        # confiugure base path
        if not spawner.environment:
            spawner.environment = {}
        spawner.environment['SUBFOLDER'] = f'{config.services.api_base_path.rstrip("/")}/north/user/{user.name}/'
        spawner.environment['JUPYTERHUB_CLIENT_API_URL'] = f'{config.north_url()}/hub/api'

        if user.name == 'anonymous':
            return

        try:
            auth_state = await user.get_auth_state()
            if not auth_state:
                self.log.warn('Authentication state is not configured!')
                return
            access_token = auth_state['access_token']
            api_headers = {'Authorization': f'Bearer {access_token}'}

            uploads_response = requests.get(
                f'{config.api_url().rstrip("/")}/v1/uploads?is_published=false&per_page=100',
                headers=api_headers)
        except Exception as e:
            self.log.error('Cannot access Nomad API: %s', e)
            return

        if uploads_response.status_code != 200:
            self.log.error('Cannot get user uploads: %s', uploads_response.text)
            return

        volumes = {}

        def add_volume(host_path, mount_path):
            host_path = os.path.abspath(host_path)
            volumes[host_path] = mount_path

        add_volume(os.path.join(config.north.users_fs, user.name), f'/prefix/work')
        add_volume(os.path.join(config.north.shared_fs), f'/prefix/shared')

        for upload in uploads_response.json()['data']:
            if 'upload_files_server_path' in upload:
                upload_id = upload['upload_id']
                upload_server_path = upload['upload_files_server_path']
                add_volume(f'{upload_server_path}/raw', f'/prefix/uploads/{upload_id}')

        self.log.debug('Configure spawner with nomad volumes: %s', volumes)

        spawner.volumes = volumes
        spawner.nomad_username = user.name


c = get_config()  # type: ignore  # pylint: disable=undefined-variable

nomad_keycloak = f'{config.keycloak.server_url.rstrip("/")}/realms/{config.keycloak.realm_name}'

c.JupyterHub.allow_named_servers = True
c.JupyterHub.tornado_settings = {
    'headers': {
        'Access-Control-Allow-Origin': '*',  # This is temporary. TODO: Place everything behind a single origin aka nginx proxy
        'Access-Control-Allow-Headers': '*',
        'Access-Control-Allow-Methods': '*'
    },
}

c.JupyterHub.port = 9000
c.JupyterHub.base_url = f'{config.services.api_base_path.rstrip("/")}/north'
c.JupyterHub.authenticator_class = NomadAuthenticator
c.Authenticator.enable_auth_state = True
c.GenericOAuthenticator.login_service = 'keycloak'
c.GenericOAuthenticator.client_id = 'nomad_public'
c.GenericOAuthenticator.authorize_url = f'{nomad_keycloak}/protocol/openid-connect/auth'
c.GenericOAuthenticator.token_url = f'{nomad_keycloak}/protocol/openid-connect/token'
c.GenericOAuthenticator.userdata_url = f'{nomad_keycloak}/protocol/openid-connect/userinfo'
c.GenericOAuthenticator.userdata_params = {'state': 'state'}
c.GenericOAuthenticator.username_key = 'preferred_username'
c.GenericOAuthenticator.userdata_method = 'GET'
c.GenericOAuthenticator.scope = ['openid', 'profile']

c.Authenticator.auto_login = True


class NomadDockerSpawner(DockerSpawner):

    async def start(self, image=None, extra_create_kwargs=None, extra_host_config=None):
        self.log.debug(f'Configuring spawner for container {self.container_name}')
        tool = self.container_name.split('-')[-1]
        if tool in tools:
            tools[tool](self)
            self.log.debug(f'Configured spawner for {tool}')
        else:
            self.log.error(f'{tool} is not a tool, raise an error')
            raise NotImplementedError('You cannot launch non tool containers.')

        return await super().start(image, extra_create_kwargs, extra_host_config)

    def _docker(self, method, *args, **kwargs):
        if config.north.windows:
            tries = 0
            max_tries = 1
            if method == 'port':
                max_tries = 3
            while tries < max_tries:
                result = super()._docker(method, *args, **kwargs)
                if result is not None:
                    break
                import time
                time.sleep(3)
                tries += 1

            return result
        else:
            return super()._docker(method, *args, **kwargs)


# launch with docker
c.JupyterHub.spawner_class = NomadDockerSpawner
c.DockerSpawner.image = 'jupyter/base-notebook'
c.DockerSpawner.remove = True
if config.north.docker_network:
    c.DockerSpawner.network_name = config.north.docker_network
c.JupyterHub.hub_ip_connect = config.north.hub_ip_connect
if config.north.hub_ip:
    c.JupyterHub.hub_ip = config.north.hub_ip
if config.north.hub_connect_url:
    c.DockerSpawner.hub_connect_url = config.north.hub_connect_url


def configure_default(spawner: DockerSpawner):
    spawner.image = 'jupyter/base-notebook'


def create_configure_from_tool_json(tool_json):
    def configure(spawner: DockerSpawner):
        spawner.image = tool_json['image']
        if 'cmd' in tools_json:
            spawner.cmd = tools_json['cmd']

        for key, value in spawner.volumes.items():
            if value.startswith('/prefix') and 'mount_path' in tool_json:
                spawner.volumes[key] = value.replace('/prefix', tool_json['mount_path'])

    return configure


tools_json_path = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    '../dependencies/nomad-remote-tools-hub/tools.json')

with open(tools_json_path, 'rt') as f:
    tools_json = json.load(f)

tools = {
    key: create_configure_from_tool_json(value)
    for key, value in tools_json.items()
}



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import re
import numpy as np
from typing import Dict, List, Union, Any, Set
import ase.data
from ase import Atoms
from matid import SymmetryAnalyzer
import matid.geometry

from nomad import config
from nomad.units import ureg
from nomad import atomutils
from nomad.utils import deep_get
from nomad.normalizing.normalizer import Normalizer
from nomad.normalizing.method import MethodNormalizer
from nomad.normalizing.material import MaterialNormalizer
from nomad.datamodel.optimade import Species
from nomad.datamodel.metainfo.simulation.system import System, Symmetry as SystemSymmetry
from nomad.datamodel.results import (
    BandGap,
    Results,
    Material,
    Method,
    GeometryOptimization,
    Properties,
    Structures,
    Structure,
    StructureOriginal,
    StructurePrimitive,
    StructureConventional,
    StructureOptimized,
    LatticeParameters,
    WyckoffSet,
    EnergyVolumeCurve,
    BulkModulus,
    ShearModulus,
    MechanicalProperties,
    ElectronicProperties,
    VibrationalProperties,
    BandStructureElectronic,
    BandStructurePhonon,
    DOSElectronic,
    DOSPhonon,
    EnergyFreeHelmholtz,
    HeatCapacityConstantVolume
)

re_label = re.compile("^([a-zA-Z][a-zA-Z]?)[^a-zA-Z]*")
elements = set(ase.data.chemical_symbols)


def valid_array(array: Any) -> bool:
    """Checks if the given variable is a non-empty array.
    """
    return array is not None and len(array) > 0


def isint(value: Any) -> bool:
    """Checks if the given variable can be interpreted as an integer.
    """
    try:
        int(value)
        return True
    except ValueError:
        return False


class ResultsNormalizer(Normalizer):
    domain = None

    def normalize(self, logger=None) -> None:
        # Setup logger
        if logger is not None:
            self.logger = logger.bind(normalizer=self.__class__.__name__)

        results = self.entry_archive.results
        if results is None:
            results = self.entry_archive.m_create(Results)
        if results.properties is None:
            results.m_create(Properties)

        if self.section_run:
            self.normalize_run(logger=self.logger)

        for measurement in self.entry_archive.measurement:
            self.normalize_measurement(measurement)

        # Add the list of available_properties: it is a selected subset of the
        # stored properties.
        available_property_names = {
            "properties.electronic.band_structure_electronic.band_gap": "electronic.band_structure_electronic.band_gap",
            "properties.electronic.band_structure_electronic": "band_structure_electronic",
            "properties.electronic.dos_electronic": "dos_electronic",
            "properties.vibrational.dos_phonon": "dos_phonon",
            "properties.vibrational.band_structure_phonon": "band_structure_phonon",
            "properties.vibrational.energy_free_helmholtz": "energy_free_helmholtz",
            "properties.vibrational.heat_capacity_constant_volume": "heat_capacity_constant_volume",
            "properties.geometry_optimization": "geometry_optimization",
            "properties.mechanical.bulk_modulus": "bulk_modulus",
            "properties.mechanical.shear_modulus": "shear_modulus",
            "properties.mechanical.energy_volume_curve": "energy_volume_curve",
            "properties.spectroscopy.eels": "eels",
        }
        available_properties: List[str] = []
        for path, shortcut in available_property_names.items():
            if deep_get(results, *path.split(".")) is not None:
                available_properties.append(shortcut)
        results.properties.available_properties = sorted(available_properties)

    def normalize_sample(self, sample) -> None:
        results = self.entry_archive.results

        if results.material is None:
            results.material = Material()
        material = results.material
        if sample.elements and len(sample.elements) > 0:
            material.elements = sample.elements
        else:
            # Try to guess elements from sample formula or name
            if sample.chemical_formula:
                try:
                    material.elements = list(set(ase.Atoms(sample.chemical_formula).get_chemical_symbols()))
                except Exception:
                    if sample.name:
                        try:
                            material.elements = list(set(ase.Atoms(sample.name).get_chemical_symbols()))
                        except Exception:
                            pass
        if sample.chemical_formula:
            material.chemical_formula_descriptive = sample.chemical_formula

        try:
            material.elements = material.elements if material.elements else []
            atoms = None
            if material.chemical_formula_descriptive:
                try:
                    atoms = ase.Atoms(material.chemical_formula_descriptive)
                except Exception as e:
                    self.logger.warn('could not normalize formula, using elements next', exc_info=e)

            if atoms is None:
                atoms = ase.Atoms(''.join(material.elements))

            if material.elements is None or len(material.elements) == 0:
                material.elements = atoms.get_chemical_symbols()

            results.material.chemical_formula_hill = atoms.get_chemical_formula(mode='hill')
            results.material.chemical_formula_reduced = atoms.get_chemical_formula(mode='reduce')
            results.material.chemical_formula_descriptive = results.material.chemical_formula_hill
        except Exception as e:
            self.logger.warn('could not normalize material', exc_info=e)

    def normalize_measurement(self, measurement) -> None:
        results = self.entry_archive.results

        # Method
        if results.method is None:
            results.method = Method(
                method_name=measurement.method_abbreviation)

        # Sample
        if results.material is None:
            results.material = Material(elements=[])
        if len(measurement.sample) > 0:
            self.normalize_sample(measurement.sample[0])

    def normalize_run(self, logger=None) -> None:
        # Fetch different information resources from which data is gathered
        repr_system = None
        for section in self.section_run.system:
            if section.is_representative:
                repr_system = section
                break
        try:
            optimade = self.entry_archive.metadata.optimade
        except Exception:
            optimade = None

        repr_symmetry = None
        if repr_system and repr_system.symmetry:
            repr_symmetry = repr_system.symmetry[0]

        # Create the section and populate the subsections
        results = self.entry_archive.results
        properties, conv_atoms, wyckoff_sets, spg_number = self.properties(repr_system, repr_symmetry)
        results.properties = properties
        results.material = MaterialNormalizer(
            self.entry_archive,
            repr_system,
            repr_symmetry,
            spg_number,
            conv_atoms,
            wyckoff_sets,
            optimade,
            logger
        ).material()
        results.method = MethodNormalizer(self.entry_archive, repr_system, results.material, logger).method()

    def species(self, labels: List[str], atomic_numbers: List[int], struct: Structure) -> None:
        """Given a list of species labels, creates the corresponding Species
        sections in the given structure.
        """
        if labels is None or atomic_numbers is None:
            return
        species: Set[str] = set()
        for label, atomic_number in zip(labels, atomic_numbers):
            if label not in species:
                species.add(label)
                i_species = struct.m_create(Species)
                i_species.name = label
                try:
                    symbol = atomutils.chemical_symbols([atomic_number])[0]
                except ValueError:
                    self.logger.info("could not identify chemical symbol for atomic number {}".format(atomic_number))
                else:
                    i_species.chemical_symbols = [symbol]
                i_species.concentration = [1.0]

    def band_structure_electronic(self) -> Union[BandStructureElectronic, None]:
        """Returns a new section containing an electronic band structure. In
        the case of multiple valid band structures, only the latest one is
        considered.

       Band structure is reported only under the following conditions:
          - There is a non-empty array of kpoints.
          - There is a non-empty array of energies.
        """
        path = ["run", "calculation", "band_structure_electronic"]
        for bs in self.traverse_reversed(path):
            if not bs.segment:
                continue
            valid = True
            for segment in bs.segment:
                energies = segment.energies
                k_points = segment.kpoints
                if not valid_array(energies) or not valid_array(k_points):
                    valid = False
                    break
            if valid:
                # Fill band structure data to the newer, improved data layout
                bs_new = BandStructureElectronic()
                bs_new.reciprocal_cell = bs
                bs_new.segment = bs.segment
                bs_new.spin_polarized = bs_new.segment[0].energies.shape[0] > 1
                bs_new.energy_fermi = bs.energy_fermi
                for info in bs.band_gap:
                    info_new = bs_new.m_create(BandGap)
                    info_new.index = info.index
                    info_new.value = info.value
                    info_new.type = info.type
                    info_new.energy_highest_occupied = info.energy_highest_occupied
                    info_new.energy_lowest_unoccupied = info.energy_lowest_unoccupied
                return bs_new

        return None

    def dos_electronic(self) -> Union[DOSElectronic, None]:
        """Returns a reference to the section containing an electronic dos. In
        the case of multiple valid DOSes, only the latest one is reported.

       DOS is reported only under the following conditions:
          - There is a non-empty array of dos_values_normalized.
          - There is a non-empty array of dos_energies.
        """
        path = ["run", "calculation", "dos_electronic"]
        for dos in self.traverse_reversed(path):
            energies = dos.energies
            values = np.array([d.value.magnitude for d in dos.total])
            if valid_array(energies) and valid_array(values):
                dos_new = DOSElectronic()
                dos_new.energies = dos
                dos_new.total = dos.total
                n_channels = values.shape[0]
                dos_new.spin_polarized = n_channels > 1
                dos_new.energy_fermi = dos.energy_fermi
                for info in dos.band_gap:
                    info_new = dos_new.m_create(BandGap)
                    info_new.index = info.index
                    info_new.energy_highest_occupied = info.energy_highest_occupied
                    info_new.energy_lowest_unoccupied = info.energy_lowest_unoccupied
                return dos_new

        return None

    def band_structure_phonon(self) -> Union[BandStructurePhonon, None]:
        """Returns a new section containing a phonon band structure. In
        the case of multiple valid band structures, only the latest one is
        considered.

       Band structure is reported only under the following conditions:
          - There is a non-empty array of kpoints.
          - There is a non-empty array of energies.
        """
        path = ["run", "calculation", "band_structure_phonon"]
        for bs in self.traverse_reversed(path):
            if not bs.segment:
                continue
            valid = True
            for segment in bs.segment:
                energies = segment.energies
                k_points = segment.kpoints
                if not valid_array(energies) or not valid_array(k_points):
                    valid = False
                    break
            if valid:
                # Fill band structure data to the newer, improved data layout
                bs_new = BandStructurePhonon()
                bs_new.segment = bs.segment
                return bs_new

        return None

    def dos_phonon(self) -> Union[DOSPhonon, None]:
        """Returns a section containing phonon dos data. In the case of
        multiple valid data sources, only the latest one is reported.

       DOS is reported only under the following conditions:
          - There is a non-empty array of values.
          - There is a non-empty array of energies.
        """
        path = ["run", "calculation", "dos_phonon"]
        for dos in self.traverse_reversed(path):
            energies = dos.energies
            values = np.array([d.value.magnitude for d in dos.total])
            if valid_array(energies) and valid_array(values):
                dos_new = DOSPhonon()
                dos_new.energies = dos
                dos_new.total = dos.total
                return dos_new

        return None

    def energy_free_helmholtz(self) -> Union[EnergyFreeHelmholtz, None]:
        """Returns a section Helmholtz free energy data. In the case of
        multiple valid data sources, only the latest one is reported.

       Helmholtz free energy is reported only under the following conditions:
          - There is a non-empty array of temperatures.
          - There is a non-empty array of energies.
        """
        path = ["workflow", "thermodynamics"]
        for thermo_prop in self.traverse_reversed(path):
            temperatures = thermo_prop.temperature
            energies = thermo_prop.vibrational_free_energy_at_constant_volume
            if valid_array(temperatures) and valid_array(energies):
                energy_free = EnergyFreeHelmholtz()
                energy_free.energies = thermo_prop
                energy_free.temperatures = thermo_prop
                return energy_free

        return None

    def heat_capacity_constant_volume(self) -> Union[HeatCapacityConstantVolume, None]:
        """Returns a section containing heat capacity data. In the case of
        multiple valid data sources, only the latest one is reported.

       Heat capacity is reported only under the following conditions:
          - There is a non-empty array of temperatures.
          - There is a non-empty array of energies.
        """
        path = ["workflow", "thermodynamics"]
        for thermo_prop in self.traverse_reversed(path):
            temperatures = thermo_prop.temperature
            heat_capacities = thermo_prop.heat_capacity_c_v
            if valid_array(temperatures) and valid_array(heat_capacities):
                heat_cap = HeatCapacityConstantVolume()
                heat_cap.heat_capacities = thermo_prop
                heat_cap.temperatures = thermo_prop
                return heat_cap

        return None

    def geometry_optimization(self) -> Union[GeometryOptimization, None]:
        """Populates both geometry optimization methodology and calculated
        properties based on the first found geometry optimization workflow.
        """
        path = ["workflow"]
        for workflow in self.traverse_reversed(path):
            # Check validity
            if workflow.type == "geometry_optimization" and workflow.calculations_ref:

                geo_opt = GeometryOptimization()
                geo_opt_wf = workflow.geometry_optimization
                geo_opt.trajectory = workflow.calculations_ref
                system_ref = workflow.calculation_result_ref.system_ref
                structure_optimized = self.nomad_system_to_structure(StructureOptimized, system_ref)
                if structure_optimized:
                    geo_opt.structure_optimized = structure_optimized
                if geo_opt_wf is not None:
                    geo_opt.type = geo_opt_wf.type
                    geo_opt.convergence_tolerance_energy_difference = geo_opt_wf.convergence_tolerance_energy_difference
                    geo_opt.convergence_tolerance_force_maximum = geo_opt_wf.convergence_tolerance_force_maximum
                    if geo_opt_wf.energies is not None:
                        geo_opt.energies = geo_opt_wf
                    geo_opt.final_energy_difference = geo_opt_wf.final_energy_difference
                    geo_opt.final_force_maximum = geo_opt_wf.final_force_maximum
                    geo_opt.final_displacement_maximum = geo_opt_wf.final_displacement_maximum
                return geo_opt

        return None

    def properties(
            self,
            repr_system: System,
            repr_symmetry: SystemSymmetry) -> tuple:
        """Returns a populated Properties subsection."""
        properties = Properties()

        # Structures
        struct_orig = None
        struct_prim = None
        struct_conv = None
        conv_atoms = None
        wyckoff_sets = None
        spg_number = None
        if repr_system:
            original_atoms = repr_system.m_cache.get("representative_atoms")
            if original_atoms:
                prim_atoms = None
                structural_type = repr_system.type
                if structural_type == "bulk":
                    conv_atoms, prim_atoms, wyckoff_sets, spg_number = self.structures_bulk(repr_symmetry)
                elif structural_type == "2D":
                    conv_atoms, prim_atoms, wyckoff_sets, spg_number = self.structures_2d(original_atoms)
                elif structural_type == "1D":
                    conv_atoms, prim_atoms = self.structures_1d(original_atoms)

                struct_orig = self.ase_atoms_to_structure(StructureOriginal, original_atoms)
                struct_prim = self.ase_atoms_to_structure(StructurePrimitive, prim_atoms)
                wyckoff_sets_serialized = wyckoff_sets if structural_type == "bulk" else None
                struct_conv = self.ase_atoms_to_structure(StructureConventional, conv_atoms, wyckoff_sets_serialized)

        if struct_orig or struct_prim or struct_conv:
            structures = Structures()
            if struct_conv:
                structures.structure_conventional = struct_conv
            if struct_prim:
                structures.structure_primitive = struct_prim
            if struct_orig:
                structures.structure_original = struct_orig
            properties.structures = structures

        # Electronic
        bs_electronic = self.band_structure_electronic()
        dos_electronic = self.dos_electronic()
        if bs_electronic or dos_electronic:
            electronic = ElectronicProperties()
            if bs_electronic:
                electronic.band_structure_electronic = bs_electronic
            if dos_electronic:
                electronic.dos_electronic = dos_electronic
            properties.electronic = electronic

        # Vibrational
        bs_phonon = self.band_structure_phonon()
        dos_phonon = self.dos_phonon()
        energy_free = self.energy_free_helmholtz()
        heat_cap = self.heat_capacity_constant_volume()
        if bs_phonon or dos_phonon or energy_free or heat_cap:
            vibrational = VibrationalProperties()
            if dos_phonon:
                vibrational.dos_phonon = dos_phonon
            if bs_phonon:
                vibrational.band_structure_phonon = bs_phonon
            if energy_free:
                vibrational.energy_free_helmholtz = energy_free
            if heat_cap:
                vibrational.heat_capacity_constant_volume = heat_cap
            properties.vibrational = vibrational

        # Mechanical
        energy_volume_curves = self.energy_volume_curves()
        bulk_modulus = self.bulk_modulus()
        shear_modulus = self.shear_modulus()
        geometry_optimization = self.geometry_optimization()
        if energy_volume_curves or bulk_modulus or shear_modulus or geometry_optimization:
            mechanical = MechanicalProperties()
            for ev in energy_volume_curves:
                mechanical.m_add_sub_section(MechanicalProperties.energy_volume_curve, ev)
            for bm in bulk_modulus:
                mechanical.m_add_sub_section(MechanicalProperties.bulk_modulus, bm)
            for sm in shear_modulus:
                mechanical.m_add_sub_section(MechanicalProperties.shear_modulus, sm)
            properties.mechanical = mechanical

        # Geometry optimization
        properties.geometry_optimization = self.geometry_optimization()

        try:
            n_calc = len(self.section_run.calculation)
        except Exception:
            n_calc = 0
        properties.n_calculations = n_calc

        return properties, conv_atoms, wyckoff_sets, spg_number

    def wyckoff_sets(self, struct: StructureConventional, wyckoff_sets: Dict) -> None:
        """Populates the Wyckoff sets in the given structure.
        """
        for group in wyckoff_sets:
            wset = struct.m_create(WyckoffSet)
            if group.x is not None or group.y is not None or group.z is not None:
                if group.x is not None:
                    wset.x = float(group.x)
                if group.y is not None:
                    wset.y = float(group.y)
                if group.z is not None:
                    wset.z = float(group.z)
            wset.indices = group.indices
            wset.element = group.element
            wset.wyckoff_letter = group.wyckoff_letter

    def structures_bulk(self, repr_symmetry):
        """The symmetry of bulk structures has already been analyzed. Here we
        use the cached results.
        """
        conv_atoms = None
        prim_atoms = None
        wyckoff_sets = None
        spg_number = None
        if repr_symmetry:
            symmetry_analyzer = repr_symmetry.m_cache.get("symmetry_analyzer")
            if symmetry_analyzer:
                spg_number = symmetry_analyzer.get_space_group_number()
                conv_atoms = symmetry_analyzer.get_conventional_system()
                prim_atoms = symmetry_analyzer.get_primitive_system()

                # For some reason MatID seems to drop the periodicity, reintroduce it here.
                conv_atoms.set_pbc(True)
                prim_atoms.set_pbc(True)
                try:
                    wyckoff_sets = symmetry_analyzer.get_wyckoff_sets_conventional(return_parameters=True)
                except Exception:
                    self.logger.error('Error resolving Wyckoff sets.')
                    wyckoff_sets = []

        return conv_atoms, prim_atoms, wyckoff_sets, spg_number

    def structures_2d(self, original_atoms):
        conv_atoms = None
        prim_atoms = None
        wyckoff_sets = None
        spg_number = None
        try:
            # Get dimension of system by also taking into account the covalent radii
            dimensions = matid.geometry.get_dimensions(original_atoms, [True, True, True])
            basis_dimensions = np.linalg.norm(original_atoms.get_cell(), axis=1)
            gaps = basis_dimensions - dimensions
            periodicity = gaps <= config.normalize.cluster_threshold

            # If two axis are not periodic, return. This only happens if the vacuum
            # gap is not aligned with a cell vector or if the linear gap search is
            # unsufficient (the structure is "wavy" making also the gap highly
            # nonlinear).
            if sum(periodicity) != 2:
                self.logger.error("could not detect the periodic dimensions in a 2D system")
                return conv_atoms, prim_atoms, wyckoff_sets, spg_number

            # Center the system in the non-periodic direction, also taking
            # periodicity into account. The get_center_of_mass()-function in MatID
            # takes into account periodicity and can produce the correct CM unlike
            # the similar function in ASE.
            pbc_cm = matid.geometry.get_center_of_mass(original_atoms)
            cell_center = 0.5 * np.sum(original_atoms.get_cell(), axis=0)
            translation = cell_center - pbc_cm
            symm_system = original_atoms.copy()
            symm_system.translate(translation)
            symm_system.wrap()

            # Set the periodicity according to detected periodicity in order
            # for SymmetryAnalyzer to use the symmetry analysis designed for 2D
            # systems.
            symm_system.set_pbc(periodicity)
            symmetry_analyzer = SymmetryAnalyzer(
                symm_system,
                config.normalize.symmetry_tolerance,
                config.normalize.flat_dim_threshold
            )

            spg_number = symmetry_analyzer.get_space_group_number()
            wyckoff_sets = symmetry_analyzer.get_wyckoff_sets_conventional(return_parameters=False)
            conv_atoms = symmetry_analyzer.get_conventional_system()
            prim_atoms = symmetry_analyzer.get_primitive_system()

            # Reduce cell size to just fit the system in the non-periodic
            # dimensions.
            conv_atoms = atomutils.get_minimized_structure(conv_atoms)
            prim_atoms = atomutils.get_minimized_structure(prim_atoms)

            # Swap the cell axes so that the non-periodic one is always the
            # last basis (=c)
            swap_dim = 2
            for i, periodic in enumerate(conv_atoms.get_pbc()):
                if not periodic:
                    non_periodic_dim = i
                    break
            if non_periodic_dim != swap_dim:
                atomutils.swap_basis(conv_atoms, non_periodic_dim, swap_dim)
                atomutils.swap_basis(prim_atoms, non_periodic_dim, swap_dim)
        except Exception as e:
            self.logger.error(
                'could not construct a conventional system for a 2D material',
                exc_info=e
            )

        return conv_atoms, prim_atoms, wyckoff_sets, spg_number

    def structures_1d(self, original_atoms):
        conv_atoms = None
        prim_atoms = None
        try:
            # First get a symmetry analyzer and the primitive system
            symm_system = original_atoms.copy()
            symm_system.set_pbc(True)
            symmetry_analyzer = SymmetryAnalyzer(
                symm_system,
                config.normalize.symmetry_tolerance,
                config.normalize.flat_dim_threshold
            )
            prim_atoms = symmetry_analyzer.get_primitive_system()
            prim_atoms.set_pbc(True)

            # Get dimension of system by also taking into account the covalent radii
            dimensions = matid.geometry.get_dimensions(prim_atoms, [True, True, True])
            basis_dimensions = np.linalg.norm(prim_atoms.get_cell(), axis=1)
            gaps = basis_dimensions - dimensions
            periodicity = gaps <= config.normalize.cluster_threshold

            # If one axis is not periodic, return. This only happens if the vacuum
            # gap is not aligned with a cell vector.
            if sum(periodicity) != 1:
                self.logger.error("could not detect the periodic dimensions in a 1D system")
                return conv_atoms, prim_atoms

            # Translate to center of mass
            conv_atoms = prim_atoms.copy()
            pbc_cm = matid.geometry.get_center_of_mass(prim_atoms)
            cell_center = 0.5 * np.sum(conv_atoms.get_cell(), axis=0)
            translation = cell_center - pbc_cm
            translation[periodicity] = 0
            conv_atoms.translate(translation)
            conv_atoms.wrap()
            conv_atoms.set_pbc(periodicity)

            # Reduce cell size to just fit the system in the non-periodic dimensions.
            conv_atoms = atomutils.get_minimized_structure(conv_atoms)

            # Swap the cell axes so that the periodic one is always the first
            # basis (=a)
            swap_dim = 0
            for i, periodic in enumerate(periodicity):
                if periodic:
                    periodic_dim = i
                    break
            if periodic_dim != swap_dim:
                atomutils.swap_basis(conv_atoms, periodic_dim, swap_dim)

            prim_atoms = conv_atoms
        except Exception as e:
            self.logger.error(
                'could not construct a conventional system for a 1D material',
                exc_info=e
            )
        return conv_atoms, prim_atoms

    def ase_atoms_to_structure(self, structure_class, atoms: Atoms, wyckoff_sets: dict = None):
        """Returns a populated instance of the given structure class from an
        ase.Atoms-object.
        """
        if not atoms or not structure_class:
            return None
        struct = structure_class()
        struct.species_at_sites = atoms.get_chemical_symbols()
        self.species(atoms.get_chemical_symbols(), atoms.get_atomic_numbers(), struct)
        struct.cartesian_site_positions = atoms.get_positions() * ureg.angstrom
        lattice_vectors = atoms.get_cell()
        if lattice_vectors is not None:
            lattice_vectors = (lattice_vectors * ureg.angstrom).to(ureg.meter).magnitude
            struct.dimension_types = [1 if x else 0 for x in atoms.get_pbc()]
            struct.lattice_vectors = lattice_vectors
            cell_volume = atomutils.get_volume(lattice_vectors)
            struct.cell_volume = cell_volume
            if atoms.get_pbc().all() and cell_volume:
                mass = atomutils.get_summed_atomic_mass(atoms.get_atomic_numbers())
                struct.mass_density = mass / cell_volume
                struct.atomic_density = len(atoms) / cell_volume
            if wyckoff_sets:
                self.wyckoff_sets(struct, wyckoff_sets)
            struct.lattice_parameters = self.lattice_parameters(lattice_vectors)
        return struct

    def nomad_system_to_structure(self, structure_class, system: System) -> Structure:
        """Returns a populated instance of the given structure class from a
        NOMAD System-section.
        """
        if not system or not structure_class:
            return None

        struct = structure_class()
        struct.cartesian_site_positions = system.atoms.positions
        struct.species_at_sites = system.atoms.labels
        self.species(system.atoms.labels, system.atoms.species, struct)
        lattice_vectors = system.atoms.lattice_vectors
        if lattice_vectors is not None:
            lattice_vectors = lattice_vectors.magnitude
            struct.dimension_types = np.array(system.atoms.periodic).astype(int)
            struct.lattice_vectors = lattice_vectors
            cell_volume = atomutils.get_volume(lattice_vectors)
            struct.cell_volume = cell_volume
            if all(system.atoms.periodic) and cell_volume:
                if system.atoms.species is not None:
                    mass = atomutils.get_summed_atomic_mass(system.atoms.species)
                    struct.mass_density = mass / cell_volume
                struct.atomic_density = len(system.atoms.labels) / cell_volume
            struct.lattice_parameters = self.lattice_parameters(lattice_vectors)
        return struct

    def lattice_parameters(self, lattice_vectors) -> LatticeParameters:
        """Converts the given cell into LatticeParameters. Undefined angle
        values are not stored.
        """
        param_values = atomutils.cell_to_cellpar(lattice_vectors)
        params = LatticeParameters()
        params.a = float(param_values[0])
        params.b = float(param_values[1])
        params.c = float(param_values[2])
        alpha = float(param_values[3])
        params.alpha = None if np.isnan(alpha) else alpha
        beta = float(param_values[4])
        params.beta = None if np.isnan(beta) else beta
        gamma = float(param_values[5])
        params.gamma = None if np.isnan(gamma) else gamma
        return params

    def energy_volume_curves(self) -> List[EnergyVolumeCurve]:
        """Returns a list containing the found EnergyVolumeCurves.
        """
        workflows = self.entry_archive.workflow
        ev_curves = []
        for workflow in workflows:
            # Equation of state must be present
            equation_of_state = workflow.equation_of_state
            if not equation_of_state:
                continue

            # Volumes must be present
            volumes = equation_of_state.volumes
            if not valid_array(volumes):
                self.logger.warning("missing eos volumes")
                continue

            # Raw EV curve
            energies_raw = equation_of_state.energies
            if valid_array(energies_raw):
                ev_curves.append(EnergyVolumeCurve(
                    type="raw",
                    volumes=equation_of_state,
                    energies_raw=equation_of_state,
                ))
            else:
                self.logger.warning("missing eos energies")

            # Fitted EV curves
            fits = equation_of_state.eos_fit
            if not fits:
                continue
            for fit in fits:
                energies_fitted = fit.fitted_energies
                function_name = fit.function_name
                if valid_array(energies_fitted):
                    ev_curves.append(EnergyVolumeCurve(
                        type=function_name,
                        volumes=equation_of_state,
                        energies_fit=fit,
                    ))

        return ev_curves

    def bulk_modulus(self) -> List[BulkModulus]:
        """Returns a list containing the found BulkModulus.
        """
        workflows = self.entry_archive.workflow
        bulk_modulus = []
        for workflow in workflows:
            # From elastic workflow
            elastic = workflow.elastic
            if elastic:
                bulk_modulus_vrh = elastic.bulk_modulus_hill
                if bulk_modulus_vrh:
                    bulk_modulus.append(BulkModulus(
                        type="voigt_reuss_hill_average",
                        value=bulk_modulus_vrh,
                    ))
                bulk_modulus_voigt = elastic.bulk_modulus_voigt
                if bulk_modulus_voigt:
                    bulk_modulus.append(BulkModulus(
                        type="voigt_average",
                        value=bulk_modulus_voigt,
                    ))
                bulk_modulus_reuss = elastic.bulk_modulus_reuss
                if bulk_modulus_reuss:
                    bulk_modulus.append(BulkModulus(
                        type="reuss_average",
                        value=bulk_modulus_reuss,
                    ))

            # From energy-volume curve fit
            equation_of_state = workflow.equation_of_state
            if equation_of_state:
                fits = equation_of_state.eos_fit
                if not fits:
                    continue
                for fit in fits:
                    modulus = fit.bulk_modulus
                    function_name = fit.function_name
                    if modulus is not None and function_name:
                        bulk_modulus.append(BulkModulus(
                            type=function_name,
                            value=modulus,
                        ))
                    else:
                        self.logger.warning("missing eos fitted energies and/or function name")

        return bulk_modulus

    def shear_modulus(self) -> List[ShearModulus]:
        """Returns a list containing the found ShearModulus.
        """
        workflows = self.entry_archive.workflow
        shear_modulus = []
        for workflow in workflows:
            # From elastic workflow
            elastic = workflow.elastic
            if elastic:
                shear_modulus_vrh = elastic.shear_modulus_hill
                if shear_modulus_vrh:
                    shear_modulus.append(ShearModulus(
                        type="voigt_reuss_hill_average",
                        value=shear_modulus_vrh,
                    ))
                shear_modulus_voigt = elastic.shear_modulus_voigt
                if shear_modulus_voigt:
                    shear_modulus.append(ShearModulus(
                        type="voigt_average",
                        value=shear_modulus_voigt,
                    ))
                shear_modulus_reuss = elastic.shear_modulus_reuss
                if shear_modulus_reuss:
                    shear_modulus.append(ShearModulus(
                        type="reuss_average",
                        value=shear_modulus_reuss,
                    ))

        return shear_modulus

    def traverse_reversed(self, path: List[str]) -> Any:
        """Traverses the given metainfo path in reverse order. Useful in
        finding the latest reported section or value.
        """
        def traverse(root, path, i):
            sections = getattr(root, path[i])
            if isinstance(sections, list):
                for section in reversed(sections):
                    if i == len(path) - 1:
                        yield section
                    else:
                        for s in traverse(section, path, i + 1):
                            yield s
            else:
                if i == len(path) - 1:
                    yield sections
                else:
                    for s in traverse(sections, path, i + 1):
                        yield s
        for t in traverse(self.entry_archive, path, 0):
            if t is not None:
                yield t
